import logging
import re
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

from sendmail_container import FormDataRequest

from cg.apps.osticket import OsTicket
from cg.models.orders.order import OrderIn
from cg.models.orders.samples import Of1508Sample
from cg.store.models import Customer, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class TicketHandler:
    """Handle tickets in the meta orders context"""

    NEW_LINE = "<br />"

    def __init__(self, osticket_api: OsTicket, status_db: Store):
        self.osticket: OsTicket = osticket_api
        self.status_db: Store = status_db

    @staticmethod
    def parse_ticket_number(name: str) -> str | None:
        """Try to parse a ticket number from a string"""
        # detect manual ticket assignment
        ticket_match = re.fullmatch(r"#([0-9]{6})", name)
        if ticket_match:
            ticket_id = ticket_match.group(1)
            LOG.info(f"{ticket_id}: detected ticket in order name")
            return ticket_id
        LOG.info(f"Could not detected ticket number in name {name}")
        return None

    def create_ticket(
        self, order: OrderIn, user_name: str, user_mail: str, project: str
    ) -> int | None:
        """Create a ticket and return the ticket number"""
        message: str = self.create_new_ticket_header(
            message=self.create_xml_sample_list(order=order, user_name=user_name),
            order=order,
            project=project,
        )
        attachment: dict = self.create_attachment(order=order)
        ticket_nr: str | None = self.osticket.open_ticket(
            name=user_name,
            email=user_mail,
            subject=order.name,
            message=message,
            attachment=attachment,
        )
        LOG.info(f"{ticket_nr}: opened new ticket")

        return ticket_nr

    def create_attachment(self, order: OrderIn):
        return self.osticket.create_new_ticket_attachment(
            content=self.replace_empty_string_with_none(obj=order.dict()), file_name="order.json"
        )

    def create_xml_sample_list(self, order: OrderIn, user_name: str) -> str:
        message = ""
        for sample in order.samples:
            message = self.add_sample_name_to_message(message=message, sample_name=sample.name)
            message = self.add_sample_apptag_to_message(
                message=message, application=sample.application
            )
            if isinstance(sample, Of1508Sample):
                message = self.add_sample_case_name_to_message(
                    message=message, case_name=sample.family_name
                )
                message = self.add_existing_sample_info_to_message(
                    message=message, customer_id=order.customer, internal_id=sample.internal_id
                )
            message = self.add_sample_priority_to_message(message=message, priority=sample.priority)
            message = self.add_sample_comment_to_message(message=message, comment=sample.comment)

        message += self.NEW_LINE
        message = self.add_order_comment_to_message(message=message, comment=order.comment)
        message = self.add_user_name_to_message(message=message, name=user_name)
        message = self.add_customer_to_message(message=message, customer_id=order.customer)

        return message

    @staticmethod
    def create_new_ticket_header(message: str, order: OrderIn, project: str) -> str:
        return (
            f"data:text/html;charset=utf-8, New order with {len(order.samples)} {project} samples:"
            + message
        )

    @staticmethod
    def add_existing_ticket_header(message: str, order: OrderIn, project: str) -> str:
        return (
            f"A new order with {len(order.samples)} {project} samples has been connected to this ticket:"
            + message
        )

    def add_sample_name_to_message(self, message: str, sample_name: str) -> str:
        message += f"{self.NEW_LINE}{sample_name}"
        return message

    @staticmethod
    def add_sample_apptag_to_message(message: str, application: str | None) -> str:
        if application:
            message += f", application: {application}"
        return message

    @staticmethod
    def add_sample_case_name_to_message(message: str, case_name: str | None) -> str:
        if case_name:
            message += f", case: {case_name}"
        return message

    def add_existing_sample_info_to_message(
        self, message: str, customer_id: str, internal_id: str | None
    ) -> str:
        if not internal_id:
            return message

        existing_sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=internal_id)

        sample_customer = ""
        if existing_sample.customer_id != customer_id:
            sample_customer = " from " + existing_sample.customer.internal_id

        message += f" (already existing sample{sample_customer})"
        return message

    @staticmethod
    def add_sample_priority_to_message(message: str, priority: str | None) -> str:
        if priority:
            message += f", priority: {priority}"
        return message

    @staticmethod
    def add_sample_comment_to_message(message: str, comment: str | None) -> str:
        if comment:
            message += f", {comment}"
        return message

    def add_order_comment_to_message(self, message: str, comment: str | None) -> str:
        if comment:
            message += f"{self.NEW_LINE}{comment}."
        return message

    def add_user_name_to_message(self, message: str, name: str | None) -> str:
        if name:
            message += f"{self.NEW_LINE}{name}"
        return message

    def add_customer_to_message(self, message: str, customer_id: str) -> str:
        """Add customer info to message and return updated message."""
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        message += f", {customer.name} ({customer_id})"
        return message

    @classmethod
    def replace_empty_string_with_none(cls, obj: Any) -> Any:
        """Recursive function that replaces empty string in nested dicts/lists with None"""
        if obj == "":
            return None
        if isinstance(obj, dict):
            for key, item in obj.items():
                if isinstance(item, list):
                    processed_list = [
                        cls.replace_empty_string_with_none(list_item) for list_item in item
                    ]
                    obj[key] = processed_list
                else:
                    obj[key] = cls.replace_empty_string_with_none(item)
        return obj

    def connect_to_ticket(
        self, order: OrderIn, user_name: str, user_mail: str, project: str, ticket_number: str
    ) -> None:
        """Appends a new order message to the ticket selected by the customer"""
        LOG.info(f"Connecting order to ticket {ticket_number}")
        message: str = self.add_existing_ticket_header(
            message=self.create_xml_sample_list(order=order, user_name=user_name),
            order=order,
            project=project,
        )
        sender_prefix, email_server_alias = user_mail.split("@")
        temp_dir: TemporaryDirectory = self.osticket.create_connecting_ticket_attachment(
            content=self.replace_empty_string_with_none(obj=order.dict())
        )
        email_form = FormDataRequest(
            sender_prefix=sender_prefix,
            email_server_alias=email_server_alias,
            request_uri=self.osticket.email_uri,
            recipients=self.osticket.osticket_email,
            mail_title=f"[#{ticket_number}]",
            mail_body=message,
            attachments=[Path(f"{temp_dir.name}/order.json")],
        )
        email_form.submit()
        temp_dir.cleanup()



"""Test for analysis"""

import logging
from datetime import datetime

import mock
import pytest

from cg.constants import FlowCellStatus, GenePanelMasterList, Priority
from cg.constants.archiving import ArchiveLocations
from cg.constants.priority import SlurmQos
from cg.constants.sequencing import Sequencers
from cg.exc import AnalysisNotReadyError
from cg.meta.archive.archive import SpringArchiveAPI
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.models.fastq import FastqFileMeta
from cg.store.models import Case, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.mark.parametrize(
    "priority,expected_slurm_qos",
    [
        (Priority.clinical_trials, SlurmQos.NORMAL),
        (Priority.research, SlurmQos.LOW),
        (Priority.standard, SlurmQos.NORMAL),
        (Priority.priority, SlurmQos.HIGH),
        (Priority.express, SlurmQos.EXPRESS),
    ],
)
def test_get_slurm_qos_for_case(mocker, case_id: str, priority, expected_slurm_qos):
    """Test qet Quality of service (SLURM QOS) from the case priority"""

    # GIVEN a case that has a priority
    mocker.patch.object(AnalysisAPI, "get_priority_for_case")
    AnalysisAPI.get_priority_for_case.return_value = priority

    # When getting the SLURM QOS for the priority
    slurm_qos = AnalysisAPI.get_slurm_qos_for_case(AnalysisAPI, case_id=case_id)

    # THEN the expected slurm QOS should be returned
    assert slurm_qos is expected_slurm_qos


def test_gene_panels_correctly_added(customer_id):
    """Test get a correct gene panel list."""

    # GIVEN a case that has a gene panel included in the gene panel master list
    default_panels_included: list[str] = [GenePanelMasterList.get_panel_names()[0]]
    master_list: list[str] = GenePanelMasterList.get_panel_names()

    # WHEN converting the gene panels between the default and the gene_panel_master_list
    list_of_gene_panels_used: list[str] = MipAnalysisAPI.get_aggregated_panels(
        customer_id=customer_id, default_panels=set(default_panels_included)
    )

    # THEN the list_of_gene_panels_used should return all gene panels
    assert set(list_of_gene_panels_used) == set(master_list)


def test_gene_panels_not_added(customer_id):
    """Test get only OMIM-AUTO and custom gene panel list."""
    # GIVEN a case that has a gene panel that is NOT included in the gene panel master list
    default_panels_not_included: list[str] = ["PANEL_NOT_IN_GENE_PANEL_MASTER_LIST"]

    # WHEN converting the gene panels between the default and the gene_panel_master_list
    list_of_gene_panels_used: list[str] = MipAnalysisAPI.get_aggregated_panels(
        customer_id=customer_id, default_panels=set(default_panels_not_included)
    )

    # THEN the list_of_gene_panels_used should return the custom panel and OMIM-AUTO
    assert set(list_of_gene_panels_used) == set(
        default_panels_not_included
        + [GenePanelMasterList.OMIM_AUTO, GenePanelMasterList.PANELAPP_GREEN]
    )


def test_is_flow_cell_check_applicable(mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store):
    """Tests that a check for flow cells being present on disk is applicable when given a case which has no
    down-sampled nor external samples."""

    # GIVEN a case
    case: Case = analysis_store.get_cases()[0]

    # GIVEN that no samples are down-sampled nor external
    for sample in case.samples:
        assert not sample.from_sample

    # WHEN checking if a flow cell check is applicable
    # THEN the method should return True
    assert mip_analysis_api._is_flow_cell_check_applicable(case_id=case.internal_id)


def test_is_flow_cell_check_not_applicable_when_external(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store
):
    """Tests that a check for flow cells being present on disk is applicable when given a case which has no
    down-sampled nor external samples."""

    # GIVEN a case
    case: Case = analysis_store.get_cases()[0]

    # WHEN marking all of its samples as external
    for sample in case.samples:
        sample.application_version.application.is_external = True

    # WHEN checking if a flow cell check is applicable
    # THEN the method should return False
    assert not mip_analysis_api._is_flow_cell_check_applicable(case_id=case.internal_id)


def test_is_flow_cell_check_not_applicable_when_down_sampled(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store
):
    """Tests that a check for flow cells being present on disk is applicable when given a case which has no
    down-sampled nor external samples."""

    # GIVEN a case
    case: Case = analysis_store.get_cases()[0]

    # WHEN marking all of its samples as down sampled from TestSample
    for sample in case.samples:
        sample.from_sample = "TestSample"

    # WHEN checking if a flow cell check is applicable
    # THEN the method should return False
    assert not mip_analysis_api._is_flow_cell_check_applicable(case_id=case.internal_id)


def test_ensure_flow_cells_on_disk_check_not_applicable(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store, caplog
):
    """Tests that ensure_flow_cells_on_disk does not perform any action
    when is_flow_cell_check_applicable returns false."""

    # GIVEN a case
    case: Case = analysis_store.get_cases()[0]

    # WHEN _is_flow_cell_check_available returns False
    with mock.patch.object(
        AnalysisAPI,
        "_is_flow_cell_check_applicable",
        return_value=False,
    ):
        caplog.set_level("INFO")
        mip_analysis_api.ensure_flow_cells_on_disk(case.internal_id)

    # THEN a warning should be logged
    assert (
        "Flow cell check is not applicable - the case is either down sampled or external."
        in caplog.text
    )


def test_ensure_flow_cells_on_disk_does_not_request_flow_cells(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store, helpers
):
    """Tests that ensure_flow_cells_on_disk does not perform any action
    when is_flow_cell_check_applicable returns True and all flow cells are ON_DISK already."""

    # GIVEN a case
    case: Case = analysis_store.get_cases()[0]

    helpers.add_flow_cell(
        analysis_store,
        flow_cell_name="flow_cell_test",
        archived_at=datetime.now(),
        sequencer_type=Sequencers.NOVASEQ,
        samples=analysis_store.get_samples_by_case_id(case.internal_id),
        status=FlowCellStatus.ON_DISK,
        date=datetime.now(),
    )

    # WHEN _is_flow_cell_check_available returns True and the attached flow cell is ON_DISK
    with mock.patch.object(
        AnalysisAPI,
        "_is_flow_cell_check_applicable",
        return_value=True,
    ), mock.patch.object(
        Store, "request_flow_cells_for_case", return_value=None
    ) as request_checker:
        mip_analysis_api.ensure_flow_cells_on_disk(case.internal_id)

    # THEN flow cells should not be requested for the case
    assert request_checker.call_count == 0


def test_ensure_flow_cells_on_disk_does_request_flow_cells(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store, helpers
):
    """Tests that ensure_flow_cells_on_disk requests a removed flow cell
    when is_flow_cell_check_applicable returns True.."""

    # GIVEN a case with a REMOVED flow cell
    case: Case = analysis_store.get_cases()[0]
    helpers.add_flow_cell(
        analysis_store,
        flow_cell_name="flow_cell_test",
        archived_at=datetime.now(),
        sequencer_type=Sequencers.NOVASEQ,
        samples=analysis_store.get_samples_by_case_id(case.internal_id),
        status=FlowCellStatus.REMOVED,
        date=datetime.now(),
    )

    # WHEN _is_flow_cell_check_available returns True
    with mock.patch.object(
        AnalysisAPI,
        "_is_flow_cell_check_applicable",
        return_value=True,
    ):
        mip_analysis_api.ensure_flow_cells_on_disk(case.internal_id)

    # THEN the flow cell's status should be set to REQUESTED for the case
    assert analysis_store.get_flow_cell_by_name("flow_cell_test").status == FlowCellStatus.REQUESTED


def test_is_case_ready_for_analysis_true(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store, helpers
):
    """Tests that is_case_ready_for_analysis returns true for a case whose flow cells are all ON_DISK and whose
    files need no decompression nor are being decompressed currently."""

    # GIVEN a case and a flow cell with status ON_DISK
    case: Case = analysis_store.get_cases()[0]
    helpers.add_flow_cell(
        analysis_store,
        flow_cell_name="flowcell_test",
        archived_at=datetime.now(),
        sequencer_type=Sequencers.NOVASEQ,
        samples=analysis_store.get_samples_by_case_id(case.internal_id),
        status=FlowCellStatus.ON_DISK,
        date=datetime.now(),
    )

    # GIVEN that no decompression is needed nor running
    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=False
    ), mock.patch.object(PrepareFastqAPI, "is_spring_decompression_running", return_value=False):
        # WHEN running is_case_ready_for_analysis
        # THEN the result should be true
        assert mip_analysis_api.is_case_ready_for_analysis(case_id=case.internal_id)


def test_is_case_ready_for_analysis_decompression_needed(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store, helpers
):
    """Tests that is_case_ready_for_analysis returns false for a case whose flow cells are all ON_DISK but whose
    files need decompression."""

    # GIVEN a case and a flow cell
    case: Case = analysis_store.get_cases()[0]
    helpers.add_flow_cell(
        analysis_store,
        flow_cell_name="flowcell_test",
        archived_at=datetime.now(),
        sequencer_type=Sequencers.NOVASEQ,
        samples=analysis_store.get_samples_by_case_id(case.internal_id),
        status=FlowCellStatus.ON_DISK,
        date=datetime.now(),
    )

    # GIVEN that some files need to be decompressed
    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=True
    ), mock.patch.object(PrepareFastqAPI, "is_spring_decompression_running", return_value=False):
        # WHEN running is_case_ready_for_analysis
        # THEN the result should be false
        assert not mip_analysis_api.is_case_ready_for_analysis(case_id=case.internal_id)


def test_is_case_ready_for_analysis_decompression_running(
    mip_hk_store,
    analysis_store,
    helpers,
    mip_analysis_api: MipDNAAnalysisAPI,
):
    """Tests that is_case_ready_for_analysis returns false for a case whose flow cells are all ON_DISK but whose
    files are being decompressed currently."""

    # GIVEN a case and a flow cell
    case: Case = analysis_store.get_cases()[0]
    helpers.add_flow_cell(
        analysis_store,
        flow_cell_name="flowcell_test",
        archived_at=datetime.now(),
        sequencer_type=Sequencers.NOVASEQ,
        samples=analysis_store.get_samples_by_case_id(case.internal_id),
        status=FlowCellStatus.ON_DISK,
        date=datetime.now(),
    )

    # GIVEN that some files are being decompressed
    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=False
    ), mock.patch.object(PrepareFastqAPI, "is_spring_decompression_running", return_value=True):
        # WHEN running is_case_ready_for_analysis
        # THEN the result should be false
        assert not mip_analysis_api.is_case_ready_for_analysis(case_id=case.internal_id)


@pytest.mark.parametrize(
    "is_spring_decompression_needed, is_spring_decompression_running",
    [(False, False), (True, False), (False, True), (True, True)],
)
def test_prepare_fastq_files_success(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store,
    helpers,
    is_spring_decompression_needed,
    is_spring_decompression_running,
):
    """Tests that no error is thrown when running prepare_fastq_files when its flow cells are all ON_DISK,
    and no spring decompression is needed nor is running."""

    # GIVEN a case with a flow cell ON_DISK
    case: Case = analysis_store.get_cases()[0]
    helpers.add_flow_cell(
        analysis_store,
        flow_cell_name="flowcell_test",
        archived_at=datetime.now(),
        sequencer_type=Sequencers.NOVASEQ,
        samples=analysis_store.get_samples_by_case_id(case.internal_id),
        status=FlowCellStatus.ON_DISK,
        date=datetime.now(),
    )

    # GIVEN that no decompression is or running and adding the files to Housekeeper goes well
    with mock.patch.object(
        PrepareFastqAPI,
        "is_spring_decompression_needed",
        return_value=is_spring_decompression_needed,
    ), mock.patch.object(MipAnalysisAPI, "decompress_case", return_value=None), mock.patch.object(
        PrepareFastqAPI,
        "is_spring_decompression_running",
        return_value=is_spring_decompression_running,
    ), mock.patch.object(
        PrepareFastqAPI, "add_decompressed_fastq_files_to_housekeeper", return_value=None
    ):
        # WHEN running prepare_fastq_files
        if is_spring_decompression_running or is_spring_decompression_needed:
            with pytest.raises(AnalysisNotReadyError):
                mip_analysis_api.prepare_fastq_files(case_id=case.internal_id, dry_run=False)
        else:
            mip_analysis_api.prepare_fastq_files(case_id=case.internal_id, dry_run=False)


def test_prepare_fastq_files_request_miria(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store, archived_spring_file
):
    """Tests that samples' input files are requested via Miria for a Clinical customer, if files are archived."""

    # GIVEN a case belonging to a non-PDC customer with at least one archived spring file
    case: Case = analysis_store.get_cases()[0]
    case.customer.data_archive_location = ArchiveLocations.KAROLINSKA_BUCKET

    # GIVEN that at least one file is archived and not retrieved
    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=False
    ), mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_running", return_value=False
    ), mock.patch.object(
        PrepareFastqAPI, "add_decompressed_fastq_files_to_housekeeper", return_value=None
    ), mock.patch.object(
        SpringArchiveAPI, "retrieve_case"
    ) as request_submitter:
        with pytest.raises(AnalysisNotReadyError):
            # WHEN running prepare_fastq_files
            # THEN an AnalysisNotReadyError should be thrown
            mip_analysis_api.prepare_fastq_files(case_id=case.internal_id, dry_run=False)

    # THEN retrieve_samples should have been invoked
    assert request_submitter.call_count == 1


def test_prepare_fastq_files_does_not_request_miria(
    mip_analysis_api: MipDNAAnalysisAPI, analysis_store: Store, helpers: StoreHelpers
):
    """Tests that samples' input files are not requested via Miria for a Clinical customer, if no files are archived."""

    # GIVEN a case belonging to a non-PDC customer

    # GIVEN that no files have entries in the Archive table
    case: Case = analysis_store.get_cases()[0]
    case.customer.data_archive_location = ArchiveLocations.KAROLINSKA_BUCKET

    # GIVEN that all flow cells have status on disk
    helpers.add_flow_cell(
        analysis_store,
        flow_cell_name="flowcell_test",
        archived_at=datetime.now(),
        sequencer_type=Sequencers.NOVASEQ,
        samples=analysis_store.get_samples_by_case_id(case.internal_id),
        status=FlowCellStatus.ON_DISK,
        date=datetime.now(),
    )

    with mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_needed", return_value=False
    ), mock.patch.object(
        PrepareFastqAPI, "is_spring_decompression_running", return_value=False
    ), mock.patch.object(
        PrepareFastqAPI, "add_decompressed_fastq_files_to_housekeeper", return_value=None
    ):
        # WHEN running prepare_fastq_files
        # THEN no error should be raised
        mip_analysis_api.prepare_fastq_files(case_id=case.internal_id, dry_run=False)


def test_are_all_spring_files_present_true_when_all_present(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    non_archived_spring_file,
    case_id: str,
    father_sample_id: str,
    helpers: StoreHelpers,
):
    """Tests that are_all_spring_files_present returns True when no files for a case are archived."""

    # GIVEN that the only Spring file belonging to a case has no Archive entry

    # WHEN checking if all spring files are present

    # THEN the result should be True
    assert mip_analysis_api.are_all_spring_files_present(case_id)


def test_are_all_spring_files_present_false_when_none_present(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    archived_spring_file,
    case_id: str,
    sample_id: str,
):
    """Tests that are_all_spring_files_present returns False when all files for a case are archived and not retrieved."""

    # GIVEN that the only Spring file belonging to a case has an Archive entry which has retrieved_at not set

    # WHEN checking if all spring files are present

    # THEN the result should be False
    assert not mip_analysis_api.are_all_spring_files_present(case_id)


def test_ensure_files_are_present(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    archived_spring_file,
    case_id: str,
    sample_id: str,
):
    """Tests that archived files are requested via Miria after the flow cell check has been performed."""

    # GIVEN that the only Spring file belonging to a case has an Archive entry which has retrieved_at not set

    # WHEN ensuring that all files are present

    with mock.patch.object(AnalysisAPI, "ensure_flow_cells_on_disk"), mock.patch.object(
        SpringArchiveAPI,
        "retrieve_case",
    ) as request_submitter:
        mip_analysis_api.ensure_files_are_present(case_id)

    # THEN the files should have been requested
    request_submitter.assert_called()


@pytest.mark.parametrize(
    "flow_cell_status, result", [(FlowCellStatus.REMOVED, True), (FlowCellStatus.ON_DISK, False)]
)
def test_does_any_spring_file_need_to_be_retrieved_flow_cell_status(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    helpers: StoreHelpers,
    flow_cell_status: str,
    result: bool,
):
    """Tests that does_any_spring_file_need_to_be_retrieved returns True if one of
    the flow cells has status 'removed' and False if the status is instead 'ondisk'."""

    # GIVEN a case with a flow cell with provided status
    case: Case = analysis_store.get_cases()[0]
    helpers.add_flow_cell(
        analysis_store,
        flow_cell_name="flowcell_test",
        archived_at=datetime.now(),
        sequencer_type=Sequencers.NOVASEQ,
        samples=analysis_store.get_samples_by_case_id(case.internal_id),
        status=flow_cell_status,
        date=datetime.now(),
    )
    # WHEN checking if any files need to be retrieved

    # THEN the outcome should be False if the flow cell status is on disk but not otherwise
    assert mip_analysis_api.does_any_file_need_to_be_retrieved(case.internal_id) == result


def test_does_any_spring_file_need_to_be_retrieved_archived_file(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    case_id: str,
    archived_spring_file,
    non_archived_spring_file,
):
    """Tests that does_any_spring_file_need_to_be_retrieved returns true if no flow cell is removed but some
    files need to be retrieved via Miria."""

    assert mip_analysis_api.does_any_file_need_to_be_retrieved(case_id)


def test_does_any_spring_file_need_to_be_retrieved_files_present(
    mip_analysis_api: MipDNAAnalysisAPI,
    analysis_store: Store,
    case_id: str,
    non_archived_spring_file,
):
    """Tests that does_any_spring_file_need_to_be_retrieved returns true if no flow cell is removed and no
    files need to be retrieved via Miria."""

    assert mip_analysis_api.does_any_file_need_to_be_retrieved(case_id)


def test_link_fastq_files_for_sample(
    analysis_store: Store,
    caplog,
    mip_analysis_api: MipDNAAnalysisAPI,
    fastq_file_meta_raw: dict,
    mocker,
):
    caplog.set_level(logging.INFO)
    # GIVEN a case
    case: Case = analysis_store.get_cases()[0]

    # GIVEN a sample
    sample: Sample = case.links[0].sample

    with mocker.patch.object(
        AnalysisAPI,
        "gather_file_metadata_for_sample",
        return_value=[FastqFileMeta.model_validate(fastq_file_meta_raw)],
    ):
        # WHEN parsing header
        mip_analysis_api.link_fastq_files_for_sample(case=case, sample=sample)

        # THEN broadcast linking of files
        assert "Linking: " in caplog.text



"""Code for uploading to scout via CLI"""

import logging
from pathlib import Path

import click
from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.cli.upload.utils import suggest_cases_to_upload
from cg.constants import Workflow
from cg.constants.constants import DRY_RUN, FileFormat
from cg.constants.scout import ScoutCustomCaseReportTags
from cg.io.controller import WriteStream
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.mip_rna import MipRNAAnalysisAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.scout.scout_load_config import ScoutLoadConfig
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.command("scout")
@click.option(
    "-r",
    "--re-upload",
    is_flag=True,
    help="re-upload existing analysis",
)
@click.option("-p", "--print", "print_console", is_flag=True, help="print config values")
@click.argument("case_id", required=False)
@click.pass_context
def upload_to_scout(context, re_upload: bool, print_console: bool, case_id: str):
    """Upload variants from analysis to Scout."""
    status_db: Store = context.obj.status_db

    LOG.info("----------------- SCOUT -----------------------")

    if not case_id:
        suggest_cases_to_upload(status_db=status_db)
        return

    context.invoke(
        create_scout_load_config, case_id=case_id, print_console=print_console, re_upload=re_upload
    )
    if not print_console:
        context.invoke(upload_case_to_scout, case_id=case_id, re_upload=re_upload)


@click.command(name="create-scout-load-config")
@click.argument("case-id")
@click.option("-p", "--print", "print_console", is_flag=True, help="Only print config")
@click.option("-r", "--re-upload", is_flag=True, help="Overwrite existing load configs")
@click.pass_obj
def create_scout_load_config(context: CGConfig, case_id: str, print_console: bool, re_upload: bool):
    """Create a load config for a case in scout and add it to housekeeper"""

    status_db: Store = context.status_db

    LOG.info("Fetching family object")
    case_obj: Case = status_db.get_case_by_internal_id(internal_id=case_id)

    if not case_obj.analyses:
        LOG.warning(f"Could not find analyses for {case_id}")
        raise click.Abort

    context.meta_apis["upload_api"]: UploadAPI = get_upload_api(cg_config=context, case=case_obj)

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api

    LOG.info("----------------- CREATE CONFIG -----------------------")
    LOG.info("Create load config")
    try:
        scout_load_config: ScoutLoadConfig = scout_upload_api.generate_config(case_obj.analyses[0])
    except SyntaxError as error:
        LOG.warning(repr(error))
        raise click.Abort from error
    LOG.info(f"Found load config {scout_load_config}")
    root_dir: str = context.meta_apis["upload_api"].analysis_api.root
    LOG.info(f"Set root dir to {root_dir}")
    file_path: Path = Path(root_dir, case_id, "scout_load.yaml")

    if print_console:
        click.echo(
            WriteStream.write_stream_from_content(
                content=scout_load_config.dict(exclude_none=True), file_format=FileFormat.YAML
            )
        )
        LOG.info(f"Would save file to {file_path}")
        return

    if file_path.exists():
        LOG.warning(f"Scout load config {file_path} already exists")
        if re_upload:
            LOG.info("Deleting old load config")
            file_path.unlink()
        else:
            LOG.info(
                "You might remove the file and try again, consider that you might also have it in housekeeper"
            )
            raise click.Abort
    LOG.info("Saving config file to disc")
    scout_upload_api.save_config_file(upload_config=scout_load_config, file_path=file_path)

    try:
        LOG.info("Add config file to housekeeper")
        scout_upload_api.add_scout_config_to_hk(
            config_file_path=file_path, case_id=case_id, delete=re_upload
        )
    except FileExistsError as error:
        LOG.warning(f"{error}, consider removing the file from housekeeper and try again")
        raise click.Abort from error


@click.command(name="upload-case-to-scout")
@click.option(
    "-r",
    "--re-upload",
    is_flag=True,
    help="re-upload existing analysis",
)
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_case_to_scout(context: CGConfig, re_upload: bool, dry_run: bool, case_id: str):
    """Upload variants and case from analysis to Scout."""

    LOG.info("----------------- UPLOAD -----------------------")

    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    scout_api: ScoutAPI = context.scout_api

    tag_name: str = UploadScoutAPI.get_load_config_tag()
    version: Version = housekeeper_api.last_version(bundle=case_id)
    scout_config_file: File | None = housekeeper_api.get_latest_file_from_version(
        version=version, tags={tag_name}
    )

    if scout_config_file is None:
        raise FileNotFoundError(f"No scout load config was found in housekeeper for {case_id}")

    LOG.info(f"Uploading case {case_id} to scout")

    if not dry_run:
        scout_api.upload(scout_load_config=scout_config_file.full_path, force=re_upload)

    LOG.info(f"Uploaded to scout using load config {scout_config_file.full_path}")
    LOG.info("Case loaded successfully to Scout")


@click.command(name="rna-to-scout")
@DRY_RUN
@click.option("-r", "--research", is_flag=True, help="Upload research report instead of clinical")
@click.argument("case_id")
@click.pass_context
def upload_rna_to_scout(context, case_id: str, dry_run: bool, research: bool) -> None:
    """Upload an RNA case's gene fusion report and junction splice files for all samples connect via subject_id."""

    LOG.info("----------------- UPLOAD RNA TO SCOUT -----------------------")

    context.invoke(upload_rna_alignment_file_to_scout, case_id=case_id, dry_run=dry_run)
    context.invoke(upload_multiqc_to_scout, case_id=case_id, dry_run=dry_run)
    context.invoke(
        upload_rna_fusion_report_to_scout, case_id=case_id, dry_run=dry_run, research=research
    )
    context.invoke(upload_rna_junctions_to_scout, case_id=case_id, dry_run=dry_run)


@click.command(name="rna-alignment-file-to-scout")
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_rna_alignment_file_to_scout(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Upload RNA alignment file for a case to Scout."""
    LOG.info("----------------- UPLOAD RNA ALIGNMENT FILE TO SCOUT -----------------------")
    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    scout_upload_api.upload_rna_alignment_file(case_id=case_id, dry_run=dry_run)


@click.command(name="rna-fusion-report-to-scout")
@DRY_RUN
@click.option("--research", is_flag=True)
@click.argument("case_id")
@click.pass_obj
def upload_rna_fusion_report_to_scout(
    context: CGConfig, dry_run: bool, case_id: str, research: bool
) -> None:
    """Upload fusion report file for a case to Scout."""

    LOG.info("----------------- UPLOAD RNA FUSION REPORT TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    scout_upload_api.upload_fusion_report_to_scout(
        dry_run=dry_run, research=research, case_id=case_id
    )


@click.command(name="rna-junctions-to-scout")
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_rna_junctions_to_scout(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Upload RNA junctions splice files to Scout."""
    LOG.info("----------------- UPLOAD RNA JUNCTIONS TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api

    scout_upload_api.upload_rna_junctions_to_scout(dry_run=dry_run, case_id=case_id)


@click.command(name="multiqc-to-scout")
@DRY_RUN
@click.argument("case_id")
@click.pass_obj
def upload_multiqc_to_scout(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Upload multiqc report to Scout."""
    LOG.info("----------------- UPLOAD MULTIQC TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    status_db: Store = context.status_db
    case: Case = status_db.get_case_by_internal_id(internal_id=case_id)
    scout_report_type, multiqc_report = scout_upload_api.get_multiqc_html_report(
        case_id=case_id, workflow=case.data_analysis
    )
    if scout_report_type == ScoutCustomCaseReportTags.MULTIQC_RNA:
        scout_upload_api.upload_rna_report_to_dna_case_in_scout(
            dry_run=dry_run,
            rna_case_id=case_id,
            report_type=scout_report_type,
            report_file=multiqc_report,
        )
    else:
        scout_upload_api.upload_report_to_scout(
            dry_run=dry_run,
            case_id=case_id,
            report_type=scout_report_type,
            report_file=multiqc_report,
        )


def get_upload_api(case: Case, cg_config: CGConfig) -> UploadAPI:
    """Return the upload API based on the data analysis type"""

    analysis_apis: dict[Workflow, UploadAPI] = {
        Workflow.BALSAMIC: BalsamicAnalysisAPI,
        Workflow.BALSAMIC_UMI: BalsamicUmiAnalysisAPI,
        Workflow.MIP_RNA: MipRNAAnalysisAPI,
        Workflow.MIP_DNA: MipDNAAnalysisAPI,
        Workflow.RNAFUSION: RnafusionAnalysisAPI,
    }

    return UploadAPI(
        config=cg_config, analysis_api=analysis_apis.get(case.data_analysis)(cg_config)
    )



from pathlib import Path

import pytest
from _pytest.fixtures import FixtureRequest

from cg.cli.demultiplex.copy_novaseqx_demultiplex_data import get_latest_analysis_path
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.sequencing import Sequencers
from cg.exc import FlowCellError
from cg.models.demultiplex.run_parameters import RunParameters
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.utils.flow_cell import get_flow_cell_id


def test_flow_cell_id(bcl2fastq_flow_cell_dir: Path):
    """Test parsing of flow cell id."""
    # GIVEN the path to a finished flow cell run
    # GIVEN the flow cell id
    flowcell_id: str = get_flow_cell_id(bcl2fastq_flow_cell_dir.name)

    # WHEN instantiating a flow cell object
    flowcell_obj = FlowCellDirectoryData(flow_cell_path=bcl2fastq_flow_cell_dir)

    # THEN assert that the flow cell id is parsed
    assert flowcell_obj.id == flowcell_id


def test_flow_cell_position(bcl2fastq_flow_cell_dir: Path):
    """Test getting flow cell position."""
    # GIVEN the path to a finished flow cell
    # GIVEN a flow cell object
    flow_cell = FlowCellDirectoryData(flow_cell_path=bcl2fastq_flow_cell_dir)

    # WHEN fetching the flow cell position
    position = flow_cell.position

    # THEN assert it is A or B
    assert position in ["A", "B"]


def test_rta_exists(novaseq_6000_pre_1_5_kits_flow_cell: FlowCellDirectoryData):
    """Test return of RTS file."""
    # GIVEN the path to a finished flow cell
    # GIVEN a flow cell object

    # WHEN fetching the path to the RTA file
    rta_file: Path = novaseq_6000_pre_1_5_kits_flow_cell.rta_complete_path

    # THEN assert that the file exists
    assert rta_file.exists()


@pytest.mark.parametrize(
    "flow_cell_fixture",
    [
        "hiseq_x_single_index_flow_cell",
        "hiseq_x_dual_index_flow_cell",
        "hiseq_2500_dual_index_flow_cell",
        "hiseq_2500_custom_index_flow_cell",
    ],
)
def test_run_parameters_path(flow_cell_fixture: str, request: FixtureRequest):
    """Test that the run parameters file is being fetched correctly for the HiSeq flow cells."""
    # GIVEN a flow cell with a run parameters
    flow_cell: FlowCellDirectoryData = request.getfixturevalue(flow_cell_fixture)

    # WHEN getting the run parameters file name
    run_parameters_path: Path = flow_cell.run_parameters_path

    # THEN it should exist and be the expected one
    assert run_parameters_path.exists()
    assert (
        run_parameters_path.name == DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE
        or run_parameters_path.name == DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE
    )


def test_run_parameters_path_when_non_existing(tmp_flow_cell_without_run_parameters_path: Path):
    """Test that getting the path of the run parameters path fails if the file does not exist."""
    # GIVEN a flowcell object with a directory without a run parameters file
    flow_cell = FlowCellDirectoryData(flow_cell_path=tmp_flow_cell_without_run_parameters_path)

    # WHEN fetching the run parameters path
    with pytest.raises(FlowCellError) as exc:
        # THEN a FlowCellError is raised
        flow_cell.run_parameters_path
    assert "No run parameters file found in flow cell" in str(exc.value)


@pytest.mark.parametrize(
    "flow_cell_fixture, expected_sequencer",
    [
        ("hiseq_2500_custom_index_flow_cell", Sequencers.HISEQGA),
        ("hiseq_x_single_index_flow_cell", Sequencers.HISEQX),
        ("novaseq_6000_post_1_5_kits_flow_cell", Sequencers.NOVASEQ),
        ("novaseq_x_flow_cell", Sequencers.NOVASEQX),
    ],
)
def test_flow_cell_run_parameters_type(
    flow_cell_fixture: str, expected_sequencer: str, request: FixtureRequest
):
    """Test that the run parameters of the flow cell is of the expected type."""
    # GIVEN a flow cell without _run_parameters
    flow_cell: FlowCellDirectoryData = request.getfixturevalue(flow_cell_fixture)
    assert not flow_cell._run_parameters

    # WHEN creating the run parameters of the flow cell
    run_parameters: RunParameters = flow_cell.run_parameters

    # THEN the run parameters sequencer is the same as of the flow cell
    assert run_parameters.sequencer == expected_sequencer


def test_has_demultiplexing_started_locally_false(tmp_flow_cell_directory_bclconvert: Path):
    # GIVEN a flow cell without a demuxstarted.txt file
    flow_cell = FlowCellDirectoryData(tmp_flow_cell_directory_bclconvert)
    assert not Path(flow_cell.path, DemultiplexingDirsAndFiles.DEMUX_STARTED).exists()

    # WHEN checking if the flow cell has started demultiplexing
    has_demux_started: bool = flow_cell.has_demultiplexing_started_locally()

    # THEN the response should be False
    assert not has_demux_started


def test_has_demultiplexing_started_locally_true(
    tmp_flow_cell_directory_bclconvert: Path,
):
    # GIVEN a flow cell with a demuxstarted.txt file
    flow_cell = FlowCellDirectoryData(tmp_flow_cell_directory_bclconvert)
    Path(flow_cell.path, DemultiplexingDirsAndFiles.DEMUX_STARTED).touch()

    # WHEN checking if the flow cell has started demultiplexing
    has_demux_started: bool = flow_cell.has_demultiplexing_started_locally()

    # THEN the response should be True
    assert has_demux_started


def test_has_demultiplexing_started_on_sequencer_true(
    novaseqx_flow_cell_dir_with_analysis_data: Path,
):
    # GIVEN a flow cell with a BCLConvert folder
    flow_cell = FlowCellDirectoryData(novaseqx_flow_cell_dir_with_analysis_data)
    Path.mkdir(
        Path(
            flow_cell.path,
            get_latest_analysis_path(flow_cell.path),
            DemultiplexingDirsAndFiles.DATA,
            DemultiplexingDirsAndFiles.BCL_CONVERT,
        )
    )

    # WHEN checking if the flow cell has started demultiplexing
    has_demux_started: bool = flow_cell.has_demultiplexing_started_on_sequencer()

    # THEN the response should be True
    assert has_demux_started


def test_has_demultiplexing_started_on_sequencer_false(
    novaseqx_flow_cell_dir_with_analysis_data: Path,
):
    # GIVEN a flow cell without a BCLConvert folder
    flow_cell = FlowCellDirectoryData(novaseqx_flow_cell_dir_with_analysis_data)
    assert not Path(
        flow_cell.path,
        get_latest_analysis_path(flow_cell.path),
        DemultiplexingDirsAndFiles.DATA,
        DemultiplexingDirsAndFiles.BCL_CONVERT,
    ).exists()

    # WHEN checking if the flow cell has started demultiplexing
    has_demux_started: bool = flow_cell.has_demultiplexing_started_on_sequencer()

    # THEN the response should be False
    assert not has_demux_started



from pathlib import Path
from cg.constants.constants import ReadDirection, FileExtensions


def generate_concatenated_fastq_delivery_path(
    fastq_directory: Path, sample_name: str, direction: int
) -> Path:
    return Path(
        fastq_directory, f"{sample_name}_{direction}{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    )


def generate_forward_concatenated_fastq_delivery_path(
    fastq_directory: Path, sample_name: str
) -> Path:
    return generate_concatenated_fastq_delivery_path(
        fastq_directory=fastq_directory,
        sample_name=sample_name,
        direction=ReadDirection.FORWARD,
    )


def generate_reverse_concatenated_fastq_delivery_path(
    fastq_directory: Path, sample_name: str
) -> Path:
    return generate_concatenated_fastq_delivery_path(
        fastq_directory=fastq_directory,
        sample_name=sample_name,
        direction=ReadDirection.REVERSE,
    )



import logging
from typing import Callable, Type

import click
from housekeeper.store.models import Archive, File
from pydantic import BaseModel, ConfigDict

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.archiving import ArchiveLocations
from cg.exc import ArchiveJobFailedError
from cg.meta.archive.ddn.ddn_data_flow_client import DDNDataFlowClient
from cg.meta.archive.models import ArchiveHandler, FileAndSample
from cg.models.cg_config import DataFlowConfig
from cg.store.models import Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)
ARCHIVE_HANDLERS: dict[str, Type[ArchiveHandler]] = {
    ArchiveLocations.KAROLINSKA_BUCKET: DDNDataFlowClient
}


class ArchiveModels(BaseModel):
    """Model containing the necessary file and sample information."""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    file_model: Callable
    handler: ArchiveHandler


class SpringArchiveAPI:
    """Class handling the archiving of sample SPRING files to an off-premise location for long
    term storage."""

    def __init__(
        self, housekeeper_api: HousekeeperAPI, status_db: Store, data_flow_config: DataFlowConfig
    ):
        self.housekeeper_api: HousekeeperAPI = housekeeper_api
        self.status_db: Store = status_db
        self.data_flow_config: DataFlowConfig = data_flow_config

    def archive_file_to_location(
        self, file_and_sample: FileAndSample, archive_handler: ArchiveHandler
    ) -> int:
        return archive_handler.archive_file(file_and_sample=file_and_sample)

    def archive_spring_files_and_add_archives_to_housekeeper(
        self, spring_file_count_limit: int | None
    ) -> None:
        """Archives all non-archived spring files. If a limit is provided, the amount of files archived is limited
        to that amount."""
        if isinstance(spring_file_count_limit, int) and spring_file_count_limit <= 0:
            LOG.warning("Please do not provide a non-positive integer as limit - exiting.")
            return
        for archive_location in ArchiveLocations:
            self.archive_files_to_location(
                archive_location=archive_location, file_limit=spring_file_count_limit
            )

    def archive_files_to_location(self, archive_location: str, file_limit: int | None) -> None:
        """Archives up to spring file count limit number of files to the provided archive location."""
        files_to_archive: list[File] = self.housekeeper_api.get_non_archived_spring_files(
            tags=[archive_location],
            limit=file_limit,
        )
        if files_to_archive:
            files_and_samples_for_location = self.add_samples_to_files(files_to_archive)
            archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](
                self.data_flow_config
            )
            for file_and_sample in files_and_samples_for_location:
                self.archive_file(file_and_sample=file_and_sample, archive_handler=archive_handler)

        else:
            LOG.info(f"No files to archive for location {archive_location}.")

    def archive_file(self, file_and_sample: FileAndSample, archive_handler: ArchiveHandler) -> None:
        job_id: int = archive_handler.archive_file(file_and_sample)
        self.housekeeper_api.add_archives(
            files=[file_and_sample.file],
            archive_task_id=job_id,
        )

    def retrieve_case(self, case_id: str) -> None:
        """Submits jobs to retrieve any archived files belonging to the given case, and updates the Archive entries
        with the retrieval job id."""
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        files_to_retrieve: list[File] = self.get_files_to_retrieve(case)
        self.retrieve_files_from_archive_location(
            files_and_samples=self.add_samples_to_files(files=files_to_retrieve),
            archive_location=case.customer.data_archive_location,
        )

    def get_files_to_retrieve(self, case: Case) -> list[File]:
        return [
            file
            for file in self.get_archived_spring_files_for_case(case)
            if file.archive.retrieval_task_id is None
        ]

    def get_archived_spring_files_for_case(self, case: Case) -> list[File]:
        """Returns a list of archived Spring files, i.e. they have entries in the Archive table
        in Housekeeper."""
        archived_files: list[File] = []
        for link in case.links:
            archived_files += self.housekeeper_api.get_archived_files_for_bundle(
                bundle_name=link.sample.internal_id,
                tags=[SequencingFileTag.SPRING],
            )
        return archived_files

    def retrieve_files_from_archive_location(
        self, files_and_samples: list[FileAndSample], archive_location: str
    ) -> None:
        """Retrieves the archived spring files for a list of samples and sets retrieval ids in Housekeeper."""
        archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](self.data_flow_config)
        job_id: int = archive_handler.retrieve_files(files_and_samples)
        LOG.info(f"Retrieval job launched with ID {job_id}")
        self.set_archive_retrieval_task_ids(
            retrieval_task_id=job_id,
            files=[file_and_sample.file for file_and_sample in files_and_samples],
        )

    def get_archived_files_from_samples(self, samples: list[Sample]) -> list[File]:
        """Gets archived spring files from the bundles corresponding to the given list of samples."""
        files: list[File] = []
        for sample in samples:
            files.extend(
                self.housekeeper_api.get_archived_files_for_bundle(
                    bundle_name=sample.internal_id, tags=[SequencingFileTag.SPRING]
                )
            )
        return files

    def set_archive_retrieval_task_ids(self, retrieval_task_id: int, files: list[File]) -> None:
        for file in files:
            self.housekeeper_api.set_archive_retrieval_task_id(
                file_id=file.id, retrieval_task_id=retrieval_task_id
            )

    def get_sample(self, file: File) -> Sample | None:
        """Fetches the Sample corresponding to a File and logs if a Sample is not found."""
        sample: Sample | None = self.status_db.get_sample_by_internal_id(file.version.bundle.name)
        if not sample:
            LOG.warning(
                f"No sample found in status_db corresponding to sample_id {file.version.bundle.name}."
                f"Skipping archiving for corresponding file {file.path}."
            )
        return sample

    def add_samples_to_files(self, files: list[File]) -> list[FileAndSample]:
        """Fetches the Sample corresponding to each File, instantiates a FileAndSample object and
        adds it to the list which is returned."""
        files_and_samples: list[FileAndSample] = []
        for file in files:
            if sample := self.get_sample(file):
                files_and_samples.append(FileAndSample(file=file, sample=sample))
        return files_and_samples

    def update_statuses_for_ongoing_tasks(self) -> None:
        """Updates any completed jobs with a finished timestamp."""
        self.update_ongoing_archivals()
        self.update_ongoing_retrievals()

    def update_ongoing_archivals(self) -> None:
        ongoing_archivals: list[Archive] = self.housekeeper_api.get_ongoing_archivals()
        archival_ids_per_location: dict[ArchiveLocations, list[int]] = (
            self.sort_archival_ids_on_archive_location(ongoing_archivals)
        )
        for archive_location in ArchiveLocations:
            if archival_ids := archival_ids_per_location.get(archive_location):
                archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](
                    self.data_flow_config
                )
                self.update_archival_jobs_for_archive_location(
                    archive_handler=archive_handler,
                    job_ids=archival_ids,
                )

    def update_ongoing_retrievals(self) -> None:
        ongoing_retrievals: list[Archive] = self.housekeeper_api.get_ongoing_retrievals()
        retrieval_ids_per_location: dict[ArchiveLocations, list[int]] = (
            self.sort_retrieval_ids_on_archive_location(ongoing_retrievals)
        )
        for archive_location in ArchiveLocations:
            if retrieval_ids := retrieval_ids_per_location.get(archive_location):
                archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](
                    self.data_flow_config
                )
                self.update_retrieval_jobs_for_archive_location(
                    archive_handler=archive_handler,
                    job_ids=retrieval_ids,
                )

    def update_archival_jobs_for_archive_location(
        self, archive_handler: ArchiveHandler, job_ids: list[int]
    ) -> None:
        for job_id in job_ids:
            self.update_ongoing_task(
                task_id=job_id, archive_handler=archive_handler, is_archival=True
            )

    def update_retrieval_jobs_for_archive_location(
        self, archive_handler: ArchiveHandler, job_ids: list[int]
    ) -> None:
        for job_id in job_ids:
            self.update_ongoing_task(
                task_id=job_id, archive_handler=archive_handler, is_archival=False
            )

    def update_ongoing_task(
        self, task_id: int, archive_handler: ArchiveHandler, is_archival: bool
    ) -> None:
        """Fetches info on an ongoing job and updates the Archive entry in Housekeeper."""
        try:
            LOG.info(f"Fetching status for job with id {task_id}")
            is_job_done: bool = archive_handler.is_job_done(task_id)
            if is_job_done:
                LOG.info(f"Job with id {task_id} has finished, updating Archive entries.")
                if is_archival:
                    self.housekeeper_api.set_archived_at(task_id)
                else:
                    self.housekeeper_api.set_retrieved_at(task_id)
            else:
                LOG.info(f"Job with id {task_id} has not yet finished.")
        except ArchiveJobFailedError as error:
            LOG.error(error)
            if is_archival:
                LOG.warning(f"Will remove archive entries with archival task ids {task_id}")
                self.housekeeper_api.delete_archives(task_id)
            else:
                LOG.warning("Will set retrieval task id to null.")
                self.housekeeper_api.update_archive_retrieved_at(
                    old_retrieval_job_id=task_id, new_retrieval_job_id=None
                )

    def sort_archival_ids_on_archive_location(
        self, archive_entries: list[Archive]
    ) -> dict[ArchiveLocations, list[int]]:
        """Returns a dictionary with keys being ArchiveLocations and the values being the subset of the given
        archival jobs which should be archived there."""

        jobs_per_location: dict[ArchiveLocations, list[int]] = {}
        jobs_and_locations: set[tuple[int, ArchiveLocations]] = (
            self.get_unique_archival_ids_and_archive_locations(archive_entries)
        )

        for archive_location in ArchiveLocations:
            jobs_per_location[ArchiveLocations(archive_location)] = [
                job_and_location[0]
                for job_and_location in jobs_and_locations
                if job_and_location[1] == archive_location
            ]
        return jobs_per_location

    def get_unique_archival_ids_and_archive_locations(
        self, archive_entries: list[Archive]
    ) -> set[tuple[int, ArchiveLocations]]:
        ids_and_locations: set[tuple[int, ArchiveLocations]] = set()
        for archive in archive_entries:
            if location := self.get_archive_location_from_file(archive.file):
                ids_and_locations.add((archive.archiving_task_id, location))
        return ids_and_locations

    def sort_retrieval_ids_on_archive_location(
        self, archive_entries: list[Archive]
    ) -> dict[ArchiveLocations, list[int]]:
        """Returns a dictionary with keys being ArchiveLocations and the values being the subset of the given
        retrieval jobs which should be archived there."""
        jobs_per_location: dict[ArchiveLocations, list[int]] = {}
        jobs_and_locations: set[tuple[int, ArchiveLocations]] = (
            self.get_unique_retrieval_ids_and_archive_locations(archive_entries)
        )
        for archive_location in ArchiveLocations:
            jobs_per_location[ArchiveLocations(archive_location)] = [
                job_and_location[0]
                for job_and_location in jobs_and_locations
                if job_and_location[1] == archive_location
            ]
        return jobs_per_location

    def get_unique_retrieval_ids_and_archive_locations(
        self, archive_entries: list[Archive]
    ) -> set[tuple[int, ArchiveLocations]]:
        ids_and_locations: set[tuple[int, ArchiveLocations]] = set()
        for archive in archive_entries:
            if location := self.get_archive_location_from_file(archive.file):
                ids_and_locations.add((archive.retrieval_task_id, location))
        return ids_and_locations

    @staticmethod
    def is_file_archived(file: File) -> bool:
        return file.archive and file.archive.archived_at

    @staticmethod
    def get_archive_location_from_file(file: File) -> ArchiveLocations | None:
        for tag_name in [tag.name for tag in file.tags]:
            if tag_name in iter(ArchiveLocations):
                LOG.debug(f"Found archive location {tag_name}")
                return tag_name
        LOG.warning("No archive location in the file tags")
        return None

    def delete_file_from_archive_location(
        self, file_and_sample: FileAndSample, archive_location: ArchiveLocations
    ) -> None:
        archive_handler: ArchiveHandler = ARCHIVE_HANDLERS[archive_location](self.data_flow_config)
        archive_handler.delete_file(file_and_sample)

    def delete_file(self, file_path: str, dry_run: bool = False) -> None:
        """Deletes the specified file where it is archived and deletes the Housekeeper record.
        Raises:
            Click.Abort if yes is not specified or the user does not confirm the deletion."""
        file: File = self.housekeeper_api.files(path=file_path).first()
        if not self.is_file_archived(file):
            LOG.warning(f"No archived file found for file {file_path} - exiting")
            return
        archive_location: ArchiveLocations | None = self.get_archive_location_from_file(file)
        if not archive_location:
            LOG.warning("No archive location could be determined - exiting")
            return
        if dry_run:
            click.echo(f"Would have deleted file {file_path} from {archive_location}.")
            return
        file_and_sample: FileAndSample = self.add_samples_to_files([file])[0]
        self.delete_file_from_archive_location(
            file_and_sample=file_and_sample, archive_location=archive_location
        )
        self.housekeeper_api.delete_file(file.id)



import logging
import sys
from pathlib import Path

from cg.constants.demultiplexing import FASTQ_FILE_SUFFIXES
from cg.constants.symbols import UNDERSCORE

LOG = logging.getLogger(__name__)


def rename_project_directory(
    project_directory: Path, flow_cell_id: str, dry_run: bool = False
) -> None:
    """Rename a project directory by adding the prefix Project_."""
    unaligned_directory: Path = project_directory.parent
    LOG.info("Check for Dragen fastq files in project directories.")
    if dragen_fastq_files_in_project_directory(project_directory):
        LOG.debug("Only Dragen fastq files found!")
        move_dragen_fastq_files(project_directory=project_directory, dry_run=dry_run)
        if dry_run:
            LOG.info("Can't continue dry run...")
            sys.exit()

    LOG.debug(f"Rename all sample directories in {unaligned_directory}")
    for sample_dir in project_directory.iterdir():
        if sample_dir.name.startswith("Sample"):
            LOG.debug(f"{sample_dir} is already renamed")
            continue
        rename_sample_directory(
            sample_directory=sample_dir, flow_cell_id=flow_cell_id, dry_run=dry_run
        )
    new_name: str = "_".join(["Project", project_directory.name])
    new_project_path: Path = Path(unaligned_directory, new_name)
    LOG.debug(f"Rename project dir {project_directory}")
    if not dry_run:
        LOG.debug(f"Rename project dir to {new_project_path}")
        project_directory.rename(new_project_path)


def move_dragen_fastq_files(project_directory: Path, dry_run: bool = False) -> None:
    """Move Dragen fastq files into sample directories"""

    for dragen_fastq_file in project_directory.iterdir():
        LOG.debug(
            f"Derive sample name from fastq file {dragen_fastq_file}: {get_dragen_sample_name(dragen_fastq_file)}",
        )
        dragen_sample_name: str = get_dragen_sample_name(dragen_fastq_file)
        LOG.debug(f"Create sample directory {project_directory / dragen_sample_name}:")
        if not dry_run:
            (project_directory / dragen_sample_name).mkdir(exist_ok=True)
        target_directory: Path = project_directory / dragen_sample_name / dragen_fastq_file.name
        LOG.debug(f"Move fastq file into sample directory: {target_directory}")
        if not dry_run:
            dragen_fastq_file.rename(target_directory)


def rename_sample_directory(
    sample_directory: Path, flow_cell_id: str, dry_run: bool = False
) -> None:
    """Rename a sample dir and all fastq files in the sample dir.

    Renaming of the sample dir means adding Sample_ as a prefix.
    """
    project_directory: Path = sample_directory.parent
    LOG.debug(f"Renaming all fastq files in {sample_directory}")
    for fastq_file in sample_directory.iterdir():
        rename_fastq_file(fastq_file=fastq_file, flow_cell_id=flow_cell_id, dry_run=dry_run)
    new_name: str = "_".join(["Sample", sample_directory.name])
    new_sample_directory: Path = Path(project_directory, new_name)
    LOG.debug(f"Renaming sample dir {sample_directory} to {new_sample_directory}")
    if not dry_run:
        sample_directory.rename(new_sample_directory)
        LOG.debug(f"Renamed sample dir to {new_sample_directory}")


def rename_fastq_file(fastq_file: Path, flow_cell_id: str, dry_run: bool = False) -> None:
    """Rename a fastq file by appending the flowcell id as a prefix."""
    new_name: str = "_".join([flow_cell_id, fastq_file.name])
    new_file: Path = Path(fastq_file.parent, new_name)
    LOG.debug(f"Renaming fastq file {fastq_file} to {new_file}")
    if not dry_run:
        fastq_file.rename(new_file)
        LOG.debug(f"Renamed fastq file to {new_file}")


def dragen_fastq_files_in_project_directory(project_directory: Path) -> bool:
    """Checks if the project directory contains Dragen fastq files instead of sample directories"""
    return all(file_.suffixes == FASTQ_FILE_SUFFIXES for file_ in project_directory.iterdir())


def get_dragen_sample_name(dragen_fastq_file: Path) -> str:
    """Derives the sample name from a dragen fastq file"""
    return dragen_fastq_file.name.split("_")[0]



import os
import shutil
from collections import namedtuple
from datetime import datetime
from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.store.models import Case, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers

FlowCellInfo = namedtuple("FlowCellInfo", "directory name sample_internal_ids")


@pytest.fixture(name="tmp_demulitplexing_dir")
def tmp_illumina_novaseq_demulitplexing_dir(
    illumina_demultiplexed_runs_directory: Path, bcl2fastq_flow_cell_full_name: str
) -> Path:
    """Return a tmp directory in demultiplexed-runs."""
    tmp_demulitplexing_dir: Path = Path(
        illumina_demultiplexed_runs_directory, bcl2fastq_flow_cell_full_name
    )
    tmp_demulitplexing_dir.mkdir(exist_ok=True, parents=True)
    return tmp_demulitplexing_dir


@pytest.fixture(name="tmp_fastq_paths")
def temp_fastq_paths(tmp_demulitplexing_dir: Path) -> list[Path]:
    """Return a list of temporary dummy fastq paths."""
    fastqs = [
        Path(tmp_demulitplexing_dir, "fastq_1.fastq.gz"),
        Path(tmp_demulitplexing_dir, "fastq_2.fastq.gz"),
    ]
    for fastq in fastqs:
        with fastq.open("w+") as fh:
            fh.write("content")
    return fastqs


@pytest.fixture
def demultiplex_fastq_file_path() -> Path:
    """Return a path to non-existent a fastq file."""
    return Path("path/to/sample_internal_id_S1_L001_R1_001.fastq.gz")


@pytest.fixture(name="tmp_sample_sheet_path")
def tmp_samplesheet_path(tmp_demulitplexing_dir: Path) -> Path:
    """Return SampleSheet in temporary demuliplexing folder."""
    tmp_sample_sheet_path = Path(tmp_demulitplexing_dir, "SampleSheet.csv")
    with tmp_sample_sheet_path.open("w+") as fh:
        fh.write("content")
    return tmp_sample_sheet_path


@pytest.fixture(name="tmp_flow_cell_run_base_path")
def tmp_flow_cell_run_base_path(project_dir: Path, bcl2fastq_flow_cell_full_name: str) -> Path:
    """Flow cell run directory in temporary folder."""

    tmp_flow_cell_run_path: Path = Path(project_dir, "flow_cells")
    tmp_flow_cell_run_path.mkdir(exist_ok=True, parents=True)

    return tmp_flow_cell_run_path


@pytest.fixture(name="tmp_flow_cell_demux_base_path")
def tmp_flow_cell_demux_base_path(project_dir: Path, bcl2fastq_flow_cell_full_name: str) -> Path:
    """Flow cell demux directory in temporary folder."""

    tmp_flow_cell_demux_path: Path = Path(project_dir, "demultiplexed-runs")
    tmp_flow_cell_demux_path.mkdir(exist_ok=True, parents=True)

    return tmp_flow_cell_demux_path


@pytest.fixture(name="flow_cell_project_id")
def flow_cell_project_id() -> int:
    """Return flow cell run project id."""
    return 174578


@pytest.fixture(name="populated_flow_cell_store")
def populated_flow_cell_store(
    family_name: str,
    bcl2fastq_flow_cell_id: str,
    sample_id: str,
    store: Store,
    helpers: StoreHelpers,
) -> Store:
    """Populate a store with a NovaSeq flow cell."""

    populated_flow_cell_store: Store = store
    sample: Sample = helpers.add_sample(store=populated_flow_cell_store, internal_id=sample_id)
    family: Case = helpers.add_case(store=populated_flow_cell_store, internal_id=family_name)
    helpers.add_relationship(
        store=populated_flow_cell_store,
        sample=sample,
        case=family,
    )
    helpers.add_flow_cell(
        store=populated_flow_cell_store,
        flow_cell_name=bcl2fastq_flow_cell_id,
        sequencer_type="novaseq",
        samples=[sample],
    )
    return populated_flow_cell_store


@pytest.fixture(name="active_flow_cell_store")
def active_flow_cell_store(
    family_name: str,
    bcl2fastq_flow_cell_id: str,
    sample_id: str,
    base_store: Store,
    helpers: StoreHelpers,
) -> Store:
    """Populate a store with a Novaseq flow cell, with active samples on it."""
    active_flow_cell_store: Store = base_store
    sample: Sample = helpers.add_sample(store=active_flow_cell_store, internal_id=sample_id)
    family: Case = helpers.add_case(
        store=active_flow_cell_store, internal_id=family_name, action="running"
    )
    helpers.add_relationship(
        store=active_flow_cell_store,
        sample=sample,
        case=family,
    )
    helpers.add_flow_cell(
        store=active_flow_cell_store,
        flow_cell_name=bcl2fastq_flow_cell_id,
        sequencer_type="novaseq",
        samples=[sample],
    )
    return active_flow_cell_store


@pytest.fixture(name="sample_level_housekeeper_api")
def sample_level_housekeeper_api(
    bcl2fastq_flow_cell_id: str,
    real_housekeeper_api: HousekeeperAPI,
    sample_id: str,
    tmp_fastq_paths: list[Path],
    helpers,
) -> HousekeeperAPI:
    """Return a mocked Housekeeper API, containing a sample bundle with related FASTQ files."""
    sample_level_housekeeper_api = real_housekeeper_api
    bundle_data = {
        "name": sample_id,
        "created": datetime.now(),
        "version": "1.0",
        "files": [
            {"path": path.as_posix(), "tags": ["fastq", bcl2fastq_flow_cell_id], "archive": False}
            for path in tmp_fastq_paths
        ],
    }
    helpers.ensure_hk_bundle(store=sample_level_housekeeper_api, bundle_data=bundle_data)
    return sample_level_housekeeper_api


@pytest.fixture(name="flow_cell_name_housekeeper_api")
def flow_cell_name_housekeeper_api(
    bcl2fastq_flow_cell_id: str,
    real_housekeeper_api: HousekeeperAPI,
    sample_id: str,
    tmp_fastq_paths: list[Path],
    tmp_sample_sheet_path: Path,
    helpers,
) -> HousekeeperAPI:
    """Return a mocked Housekeeper API, containing a sample bundle with related FASTQ files."""
    flow_cell_housekeeper_api = real_housekeeper_api
    bundle_data = {
        "name": sample_id,
        "created": datetime.now(),
        "version": "1.0",
        "files": [
            {"path": path.as_posix(), "tags": ["fastq", bcl2fastq_flow_cell_id], "archive": False}
            for path in tmp_fastq_paths
        ],
    }
    flow_cell_bundle_data = {
        "name": bcl2fastq_flow_cell_id,
        "created": datetime.now(),
        "version": "1.0",
        "files": [
            {
                "path": tmp_sample_sheet_path.as_posix(),
                "tags": ["samplesheet", bcl2fastq_flow_cell_id],
                "archive": False,
            }
        ],
    }

    helpers.ensure_hk_bundle(store=flow_cell_housekeeper_api, bundle_data=bundle_data)
    helpers.ensure_hk_bundle(store=flow_cell_housekeeper_api, bundle_data=flow_cell_bundle_data)
    return flow_cell_housekeeper_api


@pytest.fixture(scope="session")
def base_call_file() -> Path:
    return Path("Data", "Intensities", "BaseCalls", "L001", "C1.1", "L001_1.cbcl")


@pytest.fixture(scope="session")
def inter_op_file() -> Path:
    return Path(DemultiplexingDirsAndFiles.INTER_OP, "AlignmentMetricsOut.bin")


@pytest.fixture(scope="session")
def thumbnail_file() -> Path:
    return Path("Thumbnail_Images", "L001", "C1.1", "s_1_1105_green.png")


@pytest.fixture
def lsyncd_source_directory(
    tmp_path_factory,
    novaseq_x_manifest_file: Path,
    base_call_file: Path,
    inter_op_file: Path,
    thumbnail_file: Path,
) -> Path:
    """Return a temporary directory with a manifest file and three dummy files."""
    source_directory = Path(tmp_path_factory.mktemp("source"))
    shutil.copy(novaseq_x_manifest_file, source_directory)
    for file in [base_call_file, inter_op_file, thumbnail_file]:
        full_path = Path(source_directory, file)
        full_path.parent.mkdir(parents=True)
        full_path.touch()
    return source_directory


@pytest.fixture
def lsyncd_target_directory(lsyncd_source_directory: Path, tmp_path_factory) -> Path:
    """Return a copy of the temporary source directory."""
    temp_target_directory = Path(tmp_path_factory.mktemp("tmp_target"))
    target_directory = Path(lsyncd_source_directory.parent, Path(temp_target_directory, "target"))
    shutil.copytree(lsyncd_source_directory, target_directory)
    return target_directory


def get_all_files_in_directory_tree(directory: Path) -> list[Path]:
    """Get the relative paths of all files in a directory and its subdirectories."""
    files_in_directory: list[Path] = []
    for subdir, _, files in os.walk(directory):
        subdir = Path(subdir).relative_to(directory)
        files_in_directory.extend([Path(subdir, file) for file in files])
    return files_in_directory



from pathlib import Path

from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.models.mip.mip_analysis import MipAnalysis
from cg.models.mip.mip_metrics_deliverables import MIPMetricsDeliverables


def create_mip_metrics_deliverables():
    """Get an mip_metrics_deliverables object."""
    metrics_deliverables: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML,
        file_path=Path("tests", "fixtures", "apps", "mip", "case_metrics_deliverables.yaml"),
    )
    return MIPMetricsDeliverables(**metrics_deliverables)


class MockMipAnalysis(MipAnalysisAPI):
    """Mock MIP analysis object."""

    @staticmethod
    def get_latest_metadata(family_id=None):
        """Mock get_latest_metadata."""
        # Returns: dict: parsed data
        # Define output dict
        metrics: MIPMetricsDeliverables = create_mip_metrics_deliverables()
        return MipAnalysis(
            case=family_id or "yellowhog",
            genome_build="hg19",
            sample_id_metrics=metrics.sample_id_metrics,
            mip_version="v4.0.20",
            rank_model_version="1.18",
            sample_ids=["2018-20203", "2018-20204"],
            sv_rank_model_version="1.08",
        )



"""RNAfusion delivery report API."""

from cg.constants import (
    REQUIRED_APPLICATION_FIELDS,
    REQUIRED_CASE_FIELDS,
    REQUIRED_CUSTOMER_FIELDS,
    REQUIRED_DATA_ANALYSIS_RNAFUSION_FIELDS,
    REQUIRED_REPORT_FIELDS,
    REQUIRED_SAMPLE_METADATA_RNAFUSION_FIELDS,
    REQUIRED_SAMPLE_METHODS_FIELDS,
    REQUIRED_SAMPLE_RNAFUSION_FIELDS,
    REQUIRED_SAMPLE_TIMESTAMP_FIELDS,
    RNAFUSION_REPORT_ACCREDITED_APPTAGS,
    RNAFUSION_REPORT_MINIMUM_INPUT_AMOUNT,
    Workflow,
)
from cg.constants.scout import RNAFUSION_CASE_TAGS
from cg.meta.report.field_validators import (
    get_mapped_reads_fraction,
    get_million_read_pairs,
)
from cg.meta.report.report_api import ReportAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.models.analysis import AnalysisModel
from cg.models.cg_config import CGConfig
from cg.models.report.metadata import RnafusionSampleMetadataModel
from cg.models.report.report import CaseModel, ScoutReportFiles
from cg.models.report.sample import SampleModel
from cg.models.rnafusion.rnafusion import RnafusionAnalysis, RnafusionQCMetrics
from cg.store.models import Case, Sample


class RnafusionReportAPI(ReportAPI):
    """API to create RNAfusion delivery reports."""

    def __init__(self, config: CGConfig, analysis_api: RnafusionAnalysisAPI):
        super().__init__(config=config, analysis_api=analysis_api)

    def get_sample_metadata(
        self, case: Case, sample: Sample, analysis_metadata: RnafusionAnalysis
    ) -> RnafusionSampleMetadataModel:
        """Return sample metadata to include in the report."""
        sample_metrics: RnafusionQCMetrics = analysis_metadata.sample_metrics[sample.internal_id]
        return RnafusionSampleMetadataModel(
            bias_5_3=sample_metrics.median_5prime_to_3prime_bias,
            duplicates=sample_metrics.pct_duplication,
            gc_content=sample_metrics.after_filtering_gc_content,
            input_amount=self.lims_api.get_latest_rna_input_amount(sample.internal_id),
            insert_size=None,
            insert_size_peak=None,
            mapped_reads=get_mapped_reads_fraction(
                mapped_reads=sample_metrics.read_pairs_examined * 2,
                total_reads=sample_metrics.before_filtering_total_reads,
            ),
            mean_length_r1=sample_metrics.after_filtering_read1_mean_length,
            million_read_pairs=get_million_read_pairs(
                reads=sample_metrics.before_filtering_total_reads
            ),
            mrna_bases=sample_metrics.pct_mrna_bases,
            pct_adapter=sample_metrics.pct_adapter,
            pct_surviving=sample_metrics.pct_surviving,
            q20_rate=sample_metrics.after_filtering_q20_rate,
            q30_rate=sample_metrics.after_filtering_q30_rate,
            ribosomal_bases=sample_metrics.pct_ribosomal_bases,
            rin=self.lims_api.get_sample_rin(sample.internal_id),
            uniquely_mapped_reads=sample_metrics.uniquely_mapped_percent,
        )

    @staticmethod
    def is_apptag_accredited(samples: list[SampleModel]) -> bool:
        """Return Rnafusion input application tag accreditation status."""
        return all(
            sample.application.tag in RNAFUSION_REPORT_ACCREDITED_APPTAGS for sample in samples
        )

    @staticmethod
    def is_input_amount_accredited(samples: list[SampleModel]) -> bool:
        """Return Rnafusion input amount accreditation status."""
        try:
            return all(
                float(sample.metadata.input_amount) >= RNAFUSION_REPORT_MINIMUM_INPUT_AMOUNT
                for sample in samples
            )
        except ValueError:
            return False

    def is_report_accredited(
        self, samples: list[SampleModel], analysis_metadata: AnalysisModel
    ) -> bool:
        """Return whether the Rnafusion delivery report is accredited or not."""
        is_apptag_accredited: bool = self.is_apptag_accredited(samples)
        is_input_amount_accredited: bool = self.is_input_amount_accredited(samples)
        return is_apptag_accredited and is_input_amount_accredited

    def get_template_name(self) -> str:
        """Return template name to render the delivery report."""
        return Workflow.RNAFUSION + "_report.html"

    def get_scout_uploaded_files(self, case: Case) -> ScoutReportFiles:
        """Return files that will be uploaded to Scout."""
        return ScoutReportFiles(
            vcf_fusion=self.get_scout_uploaded_file_from_hk(
                case_id=case.internal_id, scout_tag="vcf_fusion"
            )
        )

    def get_required_fields(self, case: CaseModel) -> dict:
        """Return dictionary with the delivery report required fields for Rnafusion."""
        return {
            "report": REQUIRED_REPORT_FIELDS,
            "customer": REQUIRED_CUSTOMER_FIELDS,
            "case": REQUIRED_CASE_FIELDS,
            "applications": self.get_application_required_fields(
                case=case, required_fields=REQUIRED_APPLICATION_FIELDS
            ),
            "data_analysis": REQUIRED_DATA_ANALYSIS_RNAFUSION_FIELDS,
            "samples": self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_RNAFUSION_FIELDS
            ),
            "methods": self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_METHODS_FIELDS
            ),
            "timestamps": self.get_timestamp_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_TIMESTAMP_FIELDS
            ),
            "metadata": self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_METADATA_RNAFUSION_FIELDS
            ),
        }

    def get_upload_case_tags(self) -> dict:
        """Return Balsamic UMI upload case tags."""
        return RNAFUSION_CASE_TAGS



from pathlib import Path

from pydantic.v1 import BaseModel, Field

from cg.constants.constants import Strandedness
from cg.models.analysis import AnalysisModel
from cg.models.nf_analysis import NextflowSampleSheetEntry, WorkflowParameters


class RnafusionQCMetrics(BaseModel):
    """RNAfusion QC metrics."""

    after_filtering_gc_content: float | None
    after_filtering_q20_rate: float | None
    after_filtering_q30_rate: float | None
    after_filtering_read1_mean_length: float | None
    before_filtering_total_reads: float | None
    median_5prime_to_3prime_bias: float | None
    pct_adapter: float | None
    pct_mrna_bases: float | None
    pct_ribosomal_bases: float | None
    pct_surviving: float | None
    pct_duplication: float | None
    read_pairs_examined: float | None
    uniquely_mapped_percent: float | None


class RnafusionParameters(WorkflowParameters):
    """Rnafusion parameters."""

    genomes_base: Path
    all: bool = False
    arriba: bool = True
    cram: str = "arriba,starfusion"
    fastp_trim: bool = True
    fusioncatcher: bool = True
    starfusion: bool = True
    trim_tail: int = 50
    clusterOptions: str = Field(..., alias="cluster_options")
    priority: str


class RnafusionSampleSheetEntry(NextflowSampleSheetEntry):
    """Rnafusion sample sheet model."""

    strandedness: Strandedness

    @staticmethod
    def headers() -> list[str]:
        """Return sample sheet headers."""
        return ["sample", "fastq_1", "fastq_2", "strandedness"]

    def reformat_sample_content(self) -> list[list[str]]:
        """Reformat sample sheet content as a list of list, where each list represents a line in the final file."""
        return [
            [self.name, fastq_forward_read_path, fastq_reverse_read_path, str(self.strandedness)]
            for fastq_forward_read_path, fastq_reverse_read_path in zip(
                self.fastq_forward_read_paths, self.fastq_reverse_read_paths
            )
        ]


class RnafusionAnalysis(AnalysisModel):
    """Rnafusion analysis model.

    Attributes:
        sample_metrics: retrieved QC metrics associated to a sample
    """

    sample_metrics: dict[str, RnafusionQCMetrics]



from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from cg.store.models import Case, CaseSample, Order, Sample


def filter_samples_in_case_by_internal_id(
    case_samples: Query, case_internal_id: str, **kwargs
) -> Query:
    """Return samples associated with a case."""
    return case_samples.filter(Case.internal_id == case_internal_id)


def filter_cases_with_sample_by_internal_id(case_samples: Query, sample_internal_id: str, **kwargs):
    """Return cases associated with a sample internal id."""
    return case_samples.filter(Sample.internal_id == sample_internal_id)


def filter_by_order(case_samples: Query, order_id: int, **kwargs) -> Query:
    return case_samples.join(Order, Case.orders).filter(Order.id == order_id).distinct()


def get_not_received_cases(case_samples: Query, **kwargs) -> Query:
    return case_samples.filter(Sample.received_at == None).distinct()


def get_received_cases(case_samples: Query, **kwargs) -> Query:
    not_received = (
        case_samples.filter(Sample.received_at == None).with_entities(CaseSample.case_id).subquery()
    )
    return (
        case_samples.outerjoin(not_received, CaseSample.case_id == not_received.c.case_id)
        .filter(not_received.c.case_id == None)
        .distinct()
    )


def get_not_prepared_cases(case_samples: Query, **kwargs) -> Query:
    return case_samples.filter(Sample.prepared_at == None).distinct()


def get_prepared_cases(case_samples: Query, **kwargs) -> Query:
    not_prepared = (
        case_samples.filter(Sample.prepared_at == None).with_entities(CaseSample.case_id).subquery()
    )
    return (
        case_samples.outerjoin(not_prepared, CaseSample.case_id == not_prepared.c.case_id)
        .filter(not_prepared.c.case_id == None)
        .distinct()
    )


def get_not_sequenced_cases(case_samples: Query, **kwargs) -> Query:
    return case_samples.filter(Sample.last_sequenced_at == None).distinct()


def apply_case_sample_filter(
    filter_functions: list[Callable],
    case_samples: Query,
    case_internal_id: str | None = None,
    sample_entry_id: int | None = None,
    sample_internal_id: str | None = None,
    order_id: int | None = None,
) -> Query:
    """Apply filtering functions to the sample queries and return filtered results."""

    for function in filter_functions:
        case_samples: Query = function(
            case_samples=case_samples,
            case_internal_id=case_internal_id,
            sample_entry_id=sample_entry_id,
            sample_internal_id=sample_internal_id,
            order_id=order_id,
        )
    return case_samples


class CaseSampleFilter(Enum):
    """Define CaseSample filter functions."""

    SAMPLES_IN_CASE_BY_INTERNAL_ID: Callable = filter_samples_in_case_by_internal_id
    CASES_WITH_SAMPLE_BY_INTERNAL_ID: Callable = filter_cases_with_sample_by_internal_id
    CASES_WITH_SAMPLES_NOT_RECEIVED: Callable = get_not_received_cases
    CASES_WITH_ALL_SAMPLES_RECEIVED: Callable = get_received_cases
    CASES_WITH_SAMPLES_NOT_PREPARED: Callable = get_not_prepared_cases
    CASES_WITH_ALL_SAMPLES_PREPARED: Callable = get_prepared_cases
    CASES_WITH_SAMPLES_NOT_SEQUENCED: Callable = get_not_sequenced_cases
    BY_ORDER: Callable = filter_by_order



from pathlib import Path

import pytest

from cg.meta.demultiplex.housekeeper_storage_functions import (
    add_and_include_sample_sheet_path_to_housekeeper,
)


@pytest.fixture
def tmp_flow_cell_name_malformed_sample_sheet() -> str:
    """ "Returns the name of a flow cell directory ready for demultiplexing with BCL convert.
    Contains a sample sheet with malformed headers.
    """
    return "201203_A00689_0200_AHVKJCDRXY"


@pytest.fixture(name="tmp_flow_cell_name_ready_for_demultiplexing_bcl2fastq")
def tmp_flow_cell_name_ready_for_demultiplexing_bcl2fastq() -> str:
    """Returns the name of a flow cell directory ready for demultiplexing with bcl2fastq."""
    return "211101_D00483_0615_AHLG5GDRXY"


@pytest.fixture(scope="session")
def flow_cell_name_demultiplexed_with_bcl2fastq() -> str:
    """Return the name of a flow cell that has been demultiplexed with BCL2Fastq."""
    return "HHKVCALXX"


@pytest.fixture(scope="session")
def flow_cell_directory_name_demultiplexed_with_bcl2fastq(
    flow_cell_name_demultiplexed_with_bcl2fastq: str,
) -> str:
    """Return the name of a flow cell directory that has been demultiplexed with BCL2Fastq."""
    return f"170407_ST-E00198_0209_B{flow_cell_name_demultiplexed_with_bcl2fastq}"


@pytest.fixture(scope="session")
def flow_cell_name_demultiplexed_with_bcl_convert() -> str:
    return "HY7FFDRX2"


@pytest.fixture(scope="session")
def flow_cell_directory_name_demultiplexed_with_bcl_convert(
    flow_cell_name_demultiplexed_with_bcl_convert: str,
) -> str:
    return f"230504_A00689_0804_B{flow_cell_name_demultiplexed_with_bcl_convert}"


@pytest.fixture(scope="session")
def flow_cell_name_demultiplexed_with_bcl_convert() -> str:
    return "HY7FFDRX2"


@pytest.fixture(scope="session")
def flow_cell_directory_name_demultiplexed_with_bcl_convert_flat(
    flow_cell_name_demultiplexed_with_bcl_convert: str,
):
    """Return the name of a flow cell directory that has been demultiplexed with Bcl Convert using a flat output directory structure."""
    return f"230505_A00689_0804_B{flow_cell_name_demultiplexed_with_bcl_convert}"


@pytest.fixture(scope="session")
def flow_cell_directory_name_demultiplexed_with_bcl_convert_on_sequencer(
    flow_cell_name_demultiplexed_with_bcl_convert_on_sequencer: str,
):
    """Return the name of a flow cell directory that has been demultiplexed with Bcl Convert on the NovaseqX sequencer."""
    return f"20230508_LH00188_0003_A{flow_cell_name_demultiplexed_with_bcl_convert_on_sequencer}"


@pytest.fixture(scope="session")
def flow_cell_name_demultiplexed_with_bcl_convert_on_sequencer() -> str:
    """Return the name of a flow cell directory that has been demultiplexed with Bcl Convert on the NovaseqX sequencer."""
    return "22522YLT3"


@pytest.fixture(scope="session")
def hiseq_x_single_index_flow_cell_name() -> str:
    """Return the full name of a HiSeqX flow cell with only one index."""
    return "170517_ST-E00266_0210_BHJCFFALXX"


@pytest.fixture(scope="session")
def hiseq_x_dual_index_flow_cell_name() -> str:
    """Return the full name of a HiSeqX flow cell with two indexes."""
    return "180508_ST-E00269_0269_AHL32LCCXY"


@pytest.fixture(scope="session")
def hiseq_2500_dual_index_flow_cell_name() -> str:
    """Return the full name of a HiSeq2500 flow cell with double indexes."""
    return "181005_D00410_0735_BHM2LNBCX2"


@pytest.fixture(scope="session")
def hiseq_2500_custom_index_flow_cell_name() -> str:
    """Return the full name of a HiSeq2500 flow cell with double indexes."""
    return "180509_D00450_0598_BHGYFNBCX2"


@pytest.fixture(scope="session")
def bcl2fastq_flow_cell_full_name() -> str:
    """Return full flow cell name."""
    return "201203_D00483_0200_AHVKJCDRXX"


@pytest.fixture(scope="session")
def bcl_convert_flow_cell_full_name() -> str:
    """Return the full name of a bcl_convert flow cell."""
    return "211101_A00187_0615_AHLG5GDRZZ"


@pytest.fixture(scope="session")
def novaseq_6000_post_1_5_kits_flow_cell_id() -> str:
    """Return the id of a NovaSeqX flow cell."""
    return "HK33MDRX3"


@pytest.fixture(scope="session")
def novaseq_6000_post_1_5_kits_flow_cell_full_name(
    novaseq_6000_post_1_5_kits_flow_cell_id: str,
) -> str:
    """Return the full name of a NovaSeqX flow cell."""
    return f"230912_A00187_1009_A{novaseq_6000_post_1_5_kits_flow_cell_id}"


@pytest.fixture(scope="session")
def novaseq_x_flow_cell_id() -> str:
    """Return the id of a NovaSeqX flow cell."""
    return "22F52TLT3"


@pytest.fixture(scope="session")
def novaseq_x_flow_cell_full_name(novaseq_x_flow_cell_id: str) -> str:
    """Return the full name of a NovaSeqX flow cell."""
    return f"20231108_LH00188_0028_B{novaseq_x_flow_cell_id}"


@pytest.fixture
def bclconvert_flow_cell_dir_name(demux_post_processing_api) -> str:
    """Return a flow cell name that has been demultiplexed with bclconvert."""
    flow_cell_dir_name = "230504_A00689_0804_BHY7FFDRX2"
    flow_cell_path = Path(demux_post_processing_api.demultiplexed_runs_dir, flow_cell_dir_name)

    add_and_include_sample_sheet_path_to_housekeeper(
        flow_cell_directory=flow_cell_path,
        flow_cell_name="HY7FFDRX2",
        hk_api=demux_post_processing_api.hk_api,
    )
    return flow_cell_dir_name


@pytest.fixture(scope="session")
def bcl2fastq_sample_sheet_file_name() -> str:
    """Return the name of a BCL2Fastq sample sheet."""
    return "SampleSheet_bcl2fastq.csv"


# Lists


@pytest.fixture(scope="session")
def bcl_convert_demultiplexed_flow_cell_sample_internal_ids() -> list[str]:
    """
    Sample id:s present in sample sheet for dummy flow cell demultiplexed with BCL Convert in
    cg/tests/fixtures/apps/demultiplexing/demultiplexed-runs/230504_A00689_0804_BHY7FFDRX2.
    """
    return ["ACC11927A2", "ACC11927A5"]


@pytest.fixture(scope="session")
def bcl2fastq_demultiplexed_flow_cell_sample_internal_ids() -> list[str]:
    """
    Sample id:s present in sample sheet for dummy flow cell demultiplexed with BCL Convert in
    cg/tests/fixtures/apps/demultiplexing/demultiplexed-runs/170407_A00689_0209_BHHKVCALXX.
    """
    return ["SVE2528A1"]



from pydantic import BaseModel

from cg.constants import Workflow
from cg.services.orders.order_status_service.dto.order_summary import OrderSummary


class Order(BaseModel):
    customer_id: str
    ticket_id: int
    order_date: str
    id: int
    is_delivered: bool
    workflow: Workflow
    summary: OrderSummary | None = None


class OrdersResponse(BaseModel):
    orders: list[Order]
    total_count: int



import datetime as dt

import pytest

from cg.constants import Workflow
from cg.constants.constants import PrepCategory
from cg.constants.subject import PhenotypeStatus
from cg.store.models import CaseSample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="store_failing_sequencing_qc")
def store_failing_sequencing_qc(
    bcl2fastq_flow_cell_id: str,
    sample_id: str,
    ticket_id: str,
    timestamp_now: dt.datetime,
    helpers,
    store: Store,
) -> Store:
    """Populate a store with a Fluffy case, with a sample that has been sequenced on two flow cells."""
    store_case = helpers.add_case(
        store=store,
        internal_id="fluffy_case",
        name="fluffy_case",
        data_analysis=Workflow.FLUFFY,
    )

    store_sample = helpers.add_sample(
        store=store,
        application_type=PrepCategory.READY_MADE_LIBRARY.value,
        customer_id="fluffy_customer",
        is_tumour=False,
        internal_id="fluffy_sample",
        reads=5,
        original_ticket=ticket_id,
        last_sequenced_at=timestamp_now,
    )

    helpers.add_flow_cell(
        store=store,
        flow_cell_name=bcl2fastq_flow_cell_id,
        samples=[store_sample],
        date=timestamp_now,
    )

    helpers.add_relationship(store=store, case=store_case, sample=store_sample)
    helpers.ensure_sample_lane_sequencing_metrics(
        store=store,
        sample_internal_id=store_sample.internal_id,
        flow_cell_name=bcl2fastq_flow_cell_id,
        flow_cell_lane_number=1,
        sample_total_reads_in_lane=5,
        sample_base_percentage_passing_q30=30,
    )
    return store


@pytest.fixture(name="store_with_analyses_for_cases")
def store_with_analyses_for_cases(
    analysis_store: Store,
    helpers: StoreHelpers,
    timestamp_now: dt.datetime,
    timestamp_yesterday: dt.datetime,
) -> Store:
    """Return a store with two analyses for two cases."""
    case_one = analysis_store.get_case_by_internal_id("yellowhog")
    case_two = helpers.add_case(analysis_store, internal_id="test_case_1")

    cases = [case_one, case_two]
    for case in cases:
        oldest_analysis = helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_yesterday,
            completed_at=timestamp_yesterday,
            uploaded_at=timestamp_yesterday,
            delivery_reported_at=None,
        )
        helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_now,
            completed_at=timestamp_now,
            uploaded_at=timestamp_now,
            delivery_reported_at=None,
        )
        sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
        link: CaseSample = analysis_store.relate_sample(
            case=oldest_analysis.case, sample=sample, status=PhenotypeStatus.UNKNOWN
        )
        analysis_store.session.add(link)

    return analysis_store



import logging
from pathlib import Path
from typing import Type

import pytest

from cg.apps.demultiplex.sample_sheet.read_sample_sheet import (
    get_raw_samples_from_content,
    get_sample_type_from_content,
    get_samples_by_lane,
    validate_samples_are_unique,
)
from cg.apps.demultiplex.sample_sheet.sample_models import (
    FlowCellSample,
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.constants.constants import FileFormat
from cg.exc import SampleSheetError
from cg.io.controller import ReadFile


def test_validate_samples_are_unique(
    novaseq6000_flow_cell_sample_1: FlowCellSampleBcl2Fastq,
    novaseq6000_flow_cell_sample_2: FlowCellSampleBcl2Fastq,
):
    """Test that validating two different samples finishes successfully."""
    # GIVEN two different NovaSeq samples
    assert novaseq6000_flow_cell_sample_1 != novaseq6000_flow_cell_sample_2

    # WHEN validating the samples
    validate_samples_are_unique(
        samples=[novaseq6000_flow_cell_sample_1, novaseq6000_flow_cell_sample_2]
    )

    # THEN no error is raised


def test_validate_samples_are_unique_when_not_unique(
    novaseq6000_flow_cell_sample_1: FlowCellSampleBcl2Fastq, caplog
):
    """Test that validating two identical samples fails."""
    # GIVEN two identical NovaSeq samples
    caplog.set_level(logging.INFO)

    # WHEN validating the samples
    with pytest.raises(SampleSheetError):
        validate_samples_are_unique(
            samples=[novaseq6000_flow_cell_sample_1, novaseq6000_flow_cell_sample_1]
        )

    # THEN a sample sheet error is raised due to the samples being identical
    assert (
        f"Sample {novaseq6000_flow_cell_sample_1.sample_id} exists multiple times in sample sheet"
        in caplog.text
    )


def test_get_samples_by_lane(
    novaseq6000_flow_cell_sample_1: FlowCellSampleBcl2Fastq,
    novaseq6000_flow_cell_sample_2: FlowCellSampleBcl2Fastq,
):
    """Test that grouping two samples with different lanes returns two groups."""
    # GIVEN two samples on two different lanes

    # WHEN getting the samples per lane
    samples_per_lane: dict[int, list[FlowCellSample]] = get_samples_by_lane(
        samples=[novaseq6000_flow_cell_sample_1, novaseq6000_flow_cell_sample_2]
    )

    # THEN the returned value is a dictionary
    assert isinstance(samples_per_lane, dict)
    # THEN the dictionary has two entries
    assert len(samples_per_lane) == 2


def test_get_raw_samples_valid_sample_sheet(valid_sample_sheet_bcl2fastq: list[list[str]]):
    """Test that getting raw samples from a valid sample sheet gets a correct list of dictionaries."""
    # GIVEN a valid sample sheet

    # WHEN getting the list of raw samples from it
    raw_samples: list[dict[str, str]] = get_raw_samples_from_content(
        sample_sheet_content=valid_sample_sheet_bcl2fastq
    )

    # THEN it returns a list with 2 dictionaries
    assert len(raw_samples) == 2
    # THEN the list contains dictionaries
    assert isinstance(raw_samples[0], dict)
    # THEN the sample contains the key "Lane"
    assert "Lane" in raw_samples[0].keys()


def test_get_raw_samples_no_header(sample_sheet_samples_no_header: list[list[str]], caplog):
    """Test that getting samples from a sample sheet without header fails."""
    # GIVEN a sample sheet without header
    caplog.set_level(logging.INFO)

    # WHEN trying to get the samples from the sample sheet
    with pytest.raises(SampleSheetError):
        get_raw_samples_from_content(sample_sheet_content=sample_sheet_samples_no_header)

    # THEN an exception is raised because of the missing header
    assert "Could not find header in sample sheet" in caplog.text


def test_get_raw_samples_no_samples(sample_sheet_bcl2fastq_data_header: list[list[str]], caplog):
    """Test that getting samples from a sample sheet without samples fails."""
    # GIVEN a sample sheet without samples
    caplog.set_level(logging.INFO)

    # WHEN trying to get the samples from the sample sheet
    with pytest.raises(SampleSheetError):
        get_raw_samples_from_content(sample_sheet_content=sample_sheet_bcl2fastq_data_header)

    # THEN an exception is raised because of the missing samples
    assert "Could not find any samples in sample sheet" in caplog.text


def test_get_sample_type_for_bcl_convert(bcl_convert_sample_sheet_path: Path):
    # GIVEN a bcl convert sample sheet path

    # WHEN getting the sample type
    content: list[list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=bcl_convert_sample_sheet_path
    )
    sample_type: Type[FlowCellSample] = get_sample_type_from_content(content)

    # THEN the sample type is FlowCellSampleBCLConvert
    assert sample_type is FlowCellSampleBCLConvert


def test_get_sample_type_for_bcl2fastq(hiseq_x_single_index_bcl2fastq_sample_sheet_path: Path):
    # GIVEN a bcl convert sample sheet path

    # WHEN getting the sample type
    content: list[list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=hiseq_x_single_index_bcl2fastq_sample_sheet_path
    )
    sample_type: Type[FlowCellSample] = get_sample_type_from_content(content)

    # THEN the sample type is FlowCellSampleBCLConvert
    assert sample_type is FlowCellSampleBcl2Fastq



"""Module for Taxprofiler Analysis API."""

import logging
from pathlib import Path

from cg.constants import Workflow
from cg.constants.nf_analysis import MULTIQC_NEXFLOW_CONFIG
from cg.constants.sequencing import SequencingPlatform
from cg.constants.symbols import EMPTY_STRING
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.taxprofiler.taxprofiler import TaxprofilerParameters, TaxprofilerSampleSheetEntry
from cg.resources import TAXPROFILER_BUNDLE_FILENAMES_PATH
from cg.store.models import CaseSample, Sample

LOG = logging.getLogger(__name__)


class TaxprofilerAnalysisAPI(NfAnalysisAPI):
    """Handles communication between Taxprofiler processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.TAXPROFILER,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.taxprofiler.root
        self.nfcore_workflow_path: str = config.taxprofiler.workflow_path
        self.conda_env: str = config.taxprofiler.conda_env
        self.conda_binary: str = config.taxprofiler.conda_binary
        self.profile: str = config.taxprofiler.profile
        self.revision: str = config.taxprofiler.revision
        self.hostremoval_reference: Path = Path(config.taxprofiler.hostremoval_reference)
        self.databases: Path = Path(config.taxprofiler.databases)
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.taxprofiler.tower_workflow
        self.account: str = config.taxprofiler.slurm.account
        self.email: str = config.taxprofiler.slurm.mail_user
        self.nextflow_binary_path: str = config.taxprofiler.binary_path
        self.compute_env_base: str = config.taxprofiler.compute_env

    @property
    def sample_sheet_headers(self) -> list[str]:
        """Headers for sample sheet."""
        return TaxprofilerSampleSheetEntry.headers()

    @property
    def is_params_appended_to_nextflow_config(self) -> bool:
        """Return True if parameters should be added into the nextflow config file instead of the params file."""
        return False

    @property
    def is_multiple_samples_allowed(self) -> bool:
        """Return whether the analysis supports multiple samples to be linked to the case."""
        return True

    @property
    def is_multiqc_pattern_search_exact(self) -> bool:
        """Only exact pattern search is allowed to collect metrics information from multiqc file."""
        return True

    def get_nextflow_config_content(self, case_id: str) -> str:
        """Return nextflow config content."""
        return MULTIQC_NEXFLOW_CONFIG

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return Taxprofiler bundle filenames path."""
        return TAXPROFILER_BUNDLE_FILENAMES_PATH

    def get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        sample_name: str = case_sample.sample.name
        fastq_forward_read_paths, fastq_reverse_read_paths = self.get_paired_read_paths(
            sample=case_sample.sample
        )
        sample_sheet_entry = TaxprofilerSampleSheetEntry(
            name=sample_name,
            run_accession=sample_name,
            instrument_platform=SequencingPlatform.ILLUMINA,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            fasta=EMPTY_STRING,
        )
        return sample_sheet_entry.reformat_sample_content()

    def get_workflow_parameters(self, case_id: str) -> TaxprofilerParameters:
        """Return Taxprofiler parameters."""
        return TaxprofilerParameters(
            cluster_options=f"--qos={self.get_slurm_qos_for_case(case_id=case_id)}",
            input=self.get_sample_sheet_path(case_id=case_id),
            outdir=self.get_case_path(case_id=case_id),
            databases=self.databases,
            hostremoval_reference=self.hostremoval_reference,
            priority=self.account,
        )

    def get_multiqc_search_patterns(self, case_id: str) -> dict:
        """Return search patterns for MultiQC for Taxprofiler."""
        samples: list[Sample] = self.status_db.get_samples_by_case_id(case_id=case_id)
        search_patterns: dict[str, str] = {
            f"{sample.name}_{sample.name}": sample.internal_id for sample in samples
        }
        return search_patterns



""" Create a sample sheet for NovaSeq flow cells."""

import logging

from cg.apps.demultiplex.sample_sheet.read_sample_sheet import get_samples_by_lane
from cg.apps.demultiplex.sample_sheet.sample_models import (
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.constants.demultiplexing import IndexSettings, SampleSheetBCLConvertSections
from cg.models.demultiplex.run_parameters import RunParameters
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData

LOG = logging.getLogger(__name__)


class SampleSheetCreator:
    """Base class for sample sheet creation."""

    def __init__(
        self,
        flow_cell: FlowCellDirectoryData,
        lims_samples: list[FlowCellSampleBCLConvert],
    ):
        self.flow_cell: FlowCellDirectoryData = flow_cell
        self.flow_cell_id: str = flow_cell.id
        self.lims_samples: list[FlowCellSampleBCLConvert] = lims_samples
        self.run_parameters: RunParameters = flow_cell.run_parameters
        self.index_settings: IndexSettings = self.run_parameters.index_settings

    def convert_sample_to_header_dict(
        self,
        sample: FlowCellSampleBCLConvert | FlowCellSampleBcl2Fastq,
        data_column_names: list[str],
    ) -> list[str]:
        """Convert a lims sample object to a list that corresponds to the sample sheet headers."""
        if self.run_parameters.is_single_index:
            sample_serialisation: dict = sample.model_dump(
                by_alias=True, exclude={"index2", "barcode_mismatches_2"}
            )
        else:
            sample_serialisation: dict = sample.model_dump(by_alias=True)

        return [str(sample_serialisation[column_name]) for column_name in data_column_names]

    def get_additional_sections_sample_sheet(self) -> list[list[str]]:
        """Return all sections of the sample sheet that are not the data section."""
        header_section: list[list[str]] = [
            [SampleSheetBCLConvertSections.Header.HEADER.value],
            SampleSheetBCLConvertSections.Header.file_format(),
            [SampleSheetBCLConvertSections.Header.RUN_NAME.value, self.flow_cell_id],
            [
                SampleSheetBCLConvertSections.Header.INSTRUMENT_PLATFORM_TITLE.value,
                SampleSheetBCLConvertSections.Header.instrument_platform_sequencer().get(
                    self.flow_cell.sequencer_type
                ),
            ],
            SampleSheetBCLConvertSections.Header.index_orientation_forward(),
            [SampleSheetBCLConvertSections.Header.INDEX_SETTINGS.value, self.index_settings.name],
        ]
        reads_section: list[list[str]] = [
            [SampleSheetBCLConvertSections.Reads.HEADER],
            [
                SampleSheetBCLConvertSections.Reads.READ_CYCLES_1,
                self.run_parameters.get_read_1_cycles(),
            ],
            [
                SampleSheetBCLConvertSections.Reads.READ_CYCLES_2,
                self.run_parameters.get_read_2_cycles(),
            ],
            [
                SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_1,
                self.run_parameters.get_index_1_cycles(),
            ],
        ]
        if not self.run_parameters.is_single_index:
            reads_section.append(
                [
                    SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_2,
                    self.run_parameters.get_index_2_cycles(),
                ]
            )
        settings_section: list[list[str]] = [
            [SampleSheetBCLConvertSections.Settings.HEADER],
            SampleSheetBCLConvertSections.Settings.software_version(),
            SampleSheetBCLConvertSections.Settings.fastq_compression_format(),
        ]
        return header_section + reads_section + settings_section

    def get_data_section_header_and_columns(self) -> list[list[str]]:
        """Return the header and column names of the data section of the sample sheet."""
        column_names: list[str] = SampleSheetBCLConvertSections.Data.column_names()
        if self.run_parameters.is_single_index:
            column_names.remove(SampleSheetBCLConvertSections.Data.BARCODE_MISMATCHES_2)
            column_names.remove(SampleSheetBCLConvertSections.Data.INDEX_2)
        return [
            [SampleSheetBCLConvertSections.Data.HEADER],
            column_names,
        ]

    def create_sample_sheet_content(self) -> list[list[str]]:
        """Create sample sheet content with samples."""
        LOG.info("Creating sample sheet content")
        complete_data_section: list[list[str]] = self.get_data_section_header_and_columns()
        sample_sheet_content: list[list[str]] = (
            self.get_additional_sections_sample_sheet() + complete_data_section
        )
        for sample in self.lims_samples:
            sample_sheet_content.append(
                self.convert_sample_to_header_dict(
                    sample=sample,
                    data_column_names=complete_data_section[1],
                )
            )
        return sample_sheet_content

    def process_samples_for_sample_sheet(self) -> None:
        """Remove unwanted samples and adapt remaining samples."""
        for lims_sample in self.lims_samples:
            lims_sample.process_indexes(run_parameters=self.run_parameters)
        is_reverse_complement: bool = (
            self.index_settings.are_i5_override_cycles_reverse_complemented
        )
        for lane, samples_in_lane in get_samples_by_lane(self.lims_samples).items():
            LOG.info(f"Updating barcode mismatch values for samples in lane {lane}")
            for lims_sample in samples_in_lane:
                lims_sample.update_barcode_mismatches(
                    samples_to_compare=samples_in_lane,
                    is_run_single_index=self.run_parameters.is_single_index,
                    is_reverse_complement=is_reverse_complement,
                )

    def construct_sample_sheet(self) -> list[list[str]]:
        """Construct and validate the sample sheet."""
        self.process_samples_for_sample_sheet()
        sample_sheet_content: list[list[str]] = self.create_sample_sheet_content()
        return sample_sheet_content



"""Delivery report specific command options."""

import click

from cg.constants import REPORT_SUPPORTED_WORKFLOW

ARGUMENT_CASE_ID = click.argument(
    "case_id",
    required=False,
    type=str,
)

OPTION_WORKFLOW = click.option(
    "--workflow",
    type=click.Choice(REPORT_SUPPORTED_WORKFLOW),
    help="Limit delivery report generation to a specific workflow",
)

OPTION_STARTED_AT = click.option(
    "--analysis-started-at",
    help="Retrieve analysis started at a specific date (i.e. '2020-05-28  12:00:46')",
)

OPTION_FORCE_REPORT = click.option(
    "-f",
    "--force",
    "force_report",
    is_flag=True,
    default=False,
    help="Overrule report validation",
)

OPTION_DRY_RUN = click.option(
    "-d",
    "--dry-run",
    is_flag=True,
    default=False,
    help="Print to console instead of generating an html report file",
)



"""Tests for the status_db_storage_functions module of the demultiplexing post post-processing module."""

from mock import MagicMock

from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingAPI
from cg.meta.demultiplex.status_db_storage_functions import (
    add_sequencing_metrics_to_statusdb,
    delete_sequencing_metrics_from_statusdb,
    metric_has_sample_in_statusdb,
    update_sample_read_count,
)
from cg.models.cg_config import CGConfig
from cg.store.models import Flowcell, Sample, SampleLaneSequencingMetrics
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_add_single_sequencing_metrics_entry_to_statusdb(
    store_with_sequencing_metrics: Store,
    demultiplex_context: CGConfig,
    flow_cell_name: str,
    sample_id: str,
    lane: int = 1,
):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a sequencing metrics entry
    sequencing_metrics_entry = store_with_sequencing_metrics.get_metrics_entry_by_flow_cell_name_sample_internal_id_and_lane(
        flow_cell_name=flow_cell_name, sample_internal_id=sample_id, lane=lane
    )

    # WHEN adding the sequencing metrics entry to the statusdb
    add_sequencing_metrics_to_statusdb(
        sample_lane_sequencing_metrics=[sequencing_metrics_entry],
        store=demux_post_processing_api.status_db,
    )

    # THEN the sequencing metrics entry was added to the statusdb
    assert demux_post_processing_api.status_db.get_metrics_entry_by_flow_cell_name_sample_internal_id_and_lane(
        flow_cell_name=flow_cell_name, sample_internal_id=sample_id, lane=lane
    )


def test_update_sample_read_count():
    # GIVEN a sample and a read count
    sample = Sample()
    read_count: int = 100

    # GIVEN a mocked status_db
    status_db = MagicMock()
    status_db.get_number_of_reads_for_sample_passing_q30_threshold.return_value = read_count

    # WHEN updating the sample read count
    update_sample_read_count(
        sample=sample,
        q30_threshold=0,
        store=status_db,
    )

    # THEN the reads have been updated with the read count for the sample
    assert sample.reads == read_count


def test_metric_has_sample_in_statusdb(demultiplex_context: CGConfig):
    # GIVEN a store with a sample and a sequencing metric

    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a sample internal id that does not exist in statusdb
    sample_internal_id = "does_not_exist"

    # WHEN checking if the sample exists in statusdb
    assert not metric_has_sample_in_statusdb(
        sample_internal_id=sample_internal_id, store=demux_post_processing_api.status_db
    )


def test_delete_sequencing_metrics_from_statusdb_existing_metrics(
    store_with_sequencing_metrics: Store, flow_cell_name: str
):
    # GIVEN a store with sequencing metrics
    store = store_with_sequencing_metrics

    # GIVEN that the flow cell has sequencing metrics
    metrics: list[SampleLaneSequencingMetrics] = (
        store.get_sample_lane_sequencing_metrics_by_flow_cell_name(flow_cell_name)
    )
    assert metrics

    # WHEN deleting sequencing metrics from statusdb
    delete_sequencing_metrics_from_statusdb(flow_cell_id=flow_cell_name, store=store)

    # THEN the sequencing metrics should be deleted from statusdb
    assert not store.get_sample_lane_sequencing_metrics_by_flow_cell_name(
        flow_cell_name=flow_cell_name
    )


def test_delete_sequencing_metrics_from_statusdb_no_metrics(
    store_with_sequencing_metrics: Store,
    helpers: StoreHelpers,
):
    # GIVEN a store with sequencing metrics
    store = store_with_sequencing_metrics

    # GIVEN a new flow cell with no sequencing metrics
    flow_cell: Flowcell = helpers.add_flow_cell(store=store)
    metrics: list[SampleLaneSequencingMetrics] = (
        store.get_sample_lane_sequencing_metrics_by_flow_cell_name(flow_cell.name)
    )
    assert not metrics

    # WHEN deleting sequencing metrics from statusdb
    delete_sequencing_metrics_from_statusdb(flow_cell_id=flow_cell.name, store=store)

    # THEN no errors should be raised

    # THEN the sequencing metrics are still not in statusdb
    assert not store.get_sample_lane_sequencing_metrics_by_flow_cell_name(
        flow_cell_name=flow_cell.name
    )



from pydantic import BaseModel, BeforeValidator
from typing_extensions import Annotated


def parse_case_ids(input: list[str]) -> list[str]:
    return input[0].split(",") if input else []


class DeliveryMessageRequest(BaseModel):
    case_ids: Annotated[list[str], BeforeValidator(parse_case_ids)]



"""Functions that deal with modifications of the indexes."""

import logging

from pydantic import BaseModel

from cg.utils.utils import get_hamming_distance

LOG = logging.getLogger(__name__)
DNA_COMPLEMENTS: dict[str, str] = {"A": "T", "C": "G", "G": "C", "T": "A"}
INDEX_ONE_PAD_SEQUENCE: str = "AT"
INDEX_TWO_PAD_SEQUENCE: str = "AC"
LONG_INDEX_CYCLE_NR: int = 10
MINIMUM_HAMMING_DISTANCE: int = 3
SHORT_SAMPLE_INDEX_LENGTH: int = 8


def is_dual_index(index: str) -> bool:
    """Determines if an index in the raw sample sheet is dual index or not."""
    return "-" in index


class Index(BaseModel):
    """Class that represents an index."""

    name: str
    sequence: str


def is_padding_needed(index1_cycles: int, index2_cycles: int, sample_index_length: int) -> bool:
    """Returns whether a sample needs padding or not given the sample index length.
    A sample from a NovaSeq6000 flow cell needs padding if its adapted index lengths are shorter
    than the number of index cycles reads stated in the run parameters file for both indexes.
    This happens when the sample index is 8 nucleotides long and the number of index cycles read is
    10 nucleotides long.
    """
    index_cycles: int | None = index1_cycles if index1_cycles == index2_cycles else None
    return index_cycles == LONG_INDEX_CYCLE_NR and sample_index_length == SHORT_SAMPLE_INDEX_LENGTH


def get_reverse_complement_dna_seq(dna: str) -> str:
    """Generates the reverse complement of a DNA sequence."""
    LOG.debug(f"Reverse complement string {dna}")

    return "".join(DNA_COMPLEMENTS[base] for base in reversed(dna))


def pad_index_one(index_string: str) -> str:
    """Adds bases 'AT' to index one."""
    return index_string + INDEX_ONE_PAD_SEQUENCE


def pad_index_two(index_string: str, reverse_complement: bool) -> str:
    """Adds bases to index two depending on if it should be reverse complement or not."""
    if reverse_complement:
        return INDEX_TWO_PAD_SEQUENCE + index_string
    return index_string + INDEX_TWO_PAD_SEQUENCE


def get_hamming_distance_index_1(sequence_1: str, sequence_2: str) -> int:
    """
    Get the hamming distance between two index 1 sequences.
    In the case that one sequence is longer than the other, the distance is calculated between
    the shortest sequence and the first segment of equal length of the longest sequence.
    """
    shortest_index_length: int = min(len(sequence_1), len(sequence_2))
    return get_hamming_distance(
        str_1=sequence_1[:shortest_index_length], str_2=sequence_2[:shortest_index_length]
    )


def get_hamming_distance_index_2(
    sequence_1: str, sequence_2: str, is_reverse_complement: bool
) -> int:
    """
    Get the hamming distance between two index 2 sequences.
    In the case that one sequence is longer than the other, the distance is calculated between
    the shortest sequence and the last segment of equal length of the longest sequence.
    If it does not require reverse complement, the calculation is the same as for index 1.
    """
    shortest_index_length: int = min(len(sequence_1), len(sequence_2))
    return (
        get_hamming_distance(
            str_1=sequence_1[-shortest_index_length:], str_2=sequence_2[-shortest_index_length:]
        )
        if is_reverse_complement
        else get_hamming_distance(
            str_1=sequence_1[:shortest_index_length], str_2=sequence_2[:shortest_index_length]
        )
    )



"""Hold encryption sbatch templates in variables."""

FLOW_CELL_ENCRYPT_ERROR = """
if [[ -e {pending_file_path} ]]
then
    rm -f {pending_file_path}
fi
"""

FLOW_CELL_ENCRYPT_COMMANDS = """
{symmetric_passphrase}

{asymmetrically_encrypt_passphrase}

{make_tmp_encrypt_dir}

{copy_flow_cell_dir_to_tmp}

{tar_encrypt_tmp_dir} | {parallel_gzip} | {tee} | {flow_cell_symmetric_encryption}

{flow_cell_symmetric_decryption} | {md5sum}

{diff}

{mv_passphrase_file}

{remove_pending_file}

{flag_as_complete}

{remove_tmp_encrypt_dir}
"""



from pathlib import Path

import pytest
from housekeeper.store.models import File

from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingAPI
from cg.meta.demultiplex.housekeeper_storage_functions import (
    add_and_include_sample_sheet_path_to_housekeeper,
)
from cg.models.cg_config import CGConfig
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.store.models import Sample


def test_set_dry_run(
    demultiplex_context: CGConfig,
):
    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingAPI = DemuxPostProcessingAPI(config=demultiplex_context)

    # THEN dry run should be False
    assert post_demux_api.dry_run is False

    # WHEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # THEN dry run should be True
    assert post_demux_api.dry_run is True


@pytest.mark.parametrize(
    "flow_cell_fixture, sample_ids_fixture",
    [
        ("novaseq_x_flow_cell", "selected_novaseq_x_sample_ids"),
        (
            "novaseq_6000_post_1_5_kits_flow_cell",
            "selected_novaseq_6000_post_1_5_kits_sample_ids",
        ),
    ],
    ids=["BCLConvert on sequencer", "BCLConvert flat"],
)
def test_post_processing_of_flow_cell(
    flow_cell_fixture: str,
    sample_ids_fixture: str,
    updated_demultiplex_context: CGConfig,
    updated_demux_post_processing_api: DemuxPostProcessingAPI,
    tmp_illumina_demultiplexed_flow_cells_directory: Path,
    request: pytest.FixtureRequest,
):
    """Test adding a demultiplexed flow cell to the databases with. Runs on each type of
    demultiplexing software and setting used."""

    # GIVEN a demultiplexed flow cell
    flow_cell: FlowCellDirectoryData = request.getfixturevalue(flow_cell_fixture)
    flow_cell_demultiplexing_directory: str = flow_cell.full_name
    flow_cell_name: str = flow_cell.id
    sample_internal_ids: list[str] = request.getfixturevalue(sample_ids_fixture)

    # GIVEN the sample_internal_ids are present in statusdb
    for sample_internal_id in sample_internal_ids:
        assert updated_demultiplex_context.status_db.get_sample_by_internal_id(sample_internal_id)

    # GIVEN a DemuxPostProcessing API

    # GIVEN that the sample sheet is in housekeeper
    add_and_include_sample_sheet_path_to_housekeeper(
        flow_cell_directory=flow_cell.path,
        flow_cell_name=flow_cell_name,
        hk_api=updated_demux_post_processing_api.hk_api,
    )
    assert updated_demux_post_processing_api.hk_api.get_files(
        bundle=flow_cell_name, tags=[SequencingFileTag.SAMPLE_SHEET]
    ).all()

    # WHEN post-processing the demultiplexed flow cell
    updated_demux_post_processing_api.finish_flow_cell(flow_cell_demultiplexing_directory)

    # THEN a flow cell was created in statusdb
    assert updated_demux_post_processing_api.status_db.get_flow_cell_by_name(flow_cell_name)

    # THEN sequencing metrics were created for the flow cell
    assert updated_demux_post_processing_api.status_db.get_sample_lane_sequencing_metrics_by_flow_cell_name(
        flow_cell_name=flow_cell_name
    )
    # THEN the read count was calculated for all samples in the flow cell directory
    for sample_internal_id in sample_internal_ids:
        sample: Sample = updated_demux_post_processing_api.status_db.get_sample_by_internal_id(
            sample_internal_id
        )
        assert isinstance(sample.reads, int)

    # THEN a bundle was added to Housekeeper for the flow cell
    assert updated_demux_post_processing_api.hk_api.bundle(flow_cell_name)

    # THEN a bundle was added to Housekeeper for each sample
    for sample_internal_id in sample_internal_ids:
        assert updated_demux_post_processing_api.hk_api.bundle(sample_internal_id)

    # THEN a sample sheet was added to Housekeeper
    assert updated_demux_post_processing_api.hk_api.get_files(
        tags=[SequencingFileTag.SAMPLE_SHEET],
        bundle=flow_cell_name,
    ).all()

    # THEN sample fastq files were added to Housekeeper tagged with FASTQ and the flow cell name
    for sample_internal_id in sample_internal_ids:
        assert updated_demux_post_processing_api.hk_api.get_files(
            tags=[SequencingFileTag.FASTQ, flow_cell_name],
            bundle=sample_internal_id,
        ).all()

    # THEN a delivery file was created in the flow cell directory
    delivery_path = Path(
        updated_demux_post_processing_api.demultiplexed_runs_dir,
        flow_cell_demultiplexing_directory,
        DemultiplexingDirsAndFiles.DELIVERY,
    )

    assert delivery_path.exists()


def test_get_all_demultiplexed_flow_cell_out_dirs(
    demultiplex_context: CGConfig,
    tmp_illumina_demultiplexed_flow_cells_directory,
    tmp_demultiplexed_runs_bcl2fastq_directory: Path,
):
    """Test returning all flow cell directories from the demultiplexing run directory."""
    # GIVEN a demultiplex flow cell finished output directory that exist

    # GIVEN a demultiplex context
    demux_api: DemuxPostProcessingAPI = DemuxPostProcessingAPI(config=demultiplex_context)
    demux_api.demultiplexed_runs_dir = tmp_illumina_demultiplexed_flow_cells_directory

    # WHEN calling get_all_demultiplexed_flow_cell_dirs
    demultiplexed_flow_cell_dirs: list[Path] = demux_api.get_all_demultiplexed_flow_cell_dirs()

    # THEN the demultiplexed flow cells run directories should be returned
    assert tmp_demultiplexed_runs_bcl2fastq_directory in demultiplexed_flow_cell_dirs


def test_post_processing_tracks_undetermined_fastq_files(
    updated_demux_post_processing_api: DemuxPostProcessingAPI,
    hiseq_x_single_index_flow_cell: FlowCellDirectoryData,
    selected_hiseq_x_single_index_sample_ids: list[str],
):
    # GIVEN a flow cell with undetermined fastqs in a non-pooled lane

    # GIVEN that the flow cell has the sample sheet in housekeeper
    add_and_include_sample_sheet_path_to_housekeeper(
        flow_cell_directory=hiseq_x_single_index_flow_cell.path,
        flow_cell_name=hiseq_x_single_index_flow_cell.id,
        hk_api=updated_demux_post_processing_api.hk_api,
    )

    # WHEN post processing the flow cell
    updated_demux_post_processing_api.finish_flow_cell(hiseq_x_single_index_flow_cell.full_name)

    # THEN the undetermined fastqs were stored in housekeeper
    sample_internal_id: str = selected_hiseq_x_single_index_sample_ids[0]
    fastq_files: list[File] = updated_demux_post_processing_api.hk_api.get_files(
        tags=[SequencingFileTag.FASTQ],
        bundle=sample_internal_id,
    ).all()

    undetermined_fastq_files = [file for file in fastq_files if "Undetermined" in file.path]
    assert undetermined_fastq_files


def test_sample_read_count_update_is_idempotent(
    updated_demux_post_processing_api: DemuxPostProcessingAPI,
    hiseq_x_single_index_flow_cell: FlowCellDirectoryData,
    selected_hiseq_x_single_index_sample_ids: list[str],
):
    """Test that sample read counts are the same if the flow cell is processed twice."""

    # GIVEN a demultiplexed flow cell with the sample sheet in housekeeper
    add_and_include_sample_sheet_path_to_housekeeper(
        flow_cell_directory=hiseq_x_single_index_flow_cell.path,
        flow_cell_name=hiseq_x_single_index_flow_cell.id,
        hk_api=updated_demux_post_processing_api.hk_api,
    )

    # WHEN post processing the flow cell twice
    sample_internal_id: str = selected_hiseq_x_single_index_sample_ids[0]
    updated_demux_post_processing_api.finish_flow_cell(hiseq_x_single_index_flow_cell.full_name)
    first_sample_read_count: int = (
        updated_demux_post_processing_api.status_db.get_sample_by_internal_id(
            sample_internal_id
        ).reads
    )

    updated_demux_post_processing_api.finish_flow_cell(hiseq_x_single_index_flow_cell.full_name)
    second_sample_read_count: int = (
        updated_demux_post_processing_api.status_db.get_sample_by_internal_id(
            sample_internal_id
        ).reads
    )

    # THEN the sample read counts are not zero
    assert first_sample_read_count

    # THEN the sample read count is the same after the second post processing
    assert first_sample_read_count == second_sample_read_count



"""CLI support to create config and/or start Jasen."""

import logging

import click

from cg.constants.constants import MetaApis
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.jasen import JasenAnalysisAPI

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.pass_context
def jasen(context: click.Context) -> None:
    """GMS/Jasen analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = JasenAnalysisAPI(config=context.obj)



import datetime as dt
import logging
from abc import ABC, abstractmethod

from cg.apps.lims import LimsAPI
from cg.models.orders.order import OrderIn
from cg.store.models import Base, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class Submitter(ABC):
    def __init__(self, lims: LimsAPI, status: Store):
        self.lims = lims
        self.status = status

    def validate_order(self, order: OrderIn) -> None:
        """Part of Submitter interface, base implementation"""

    @abstractmethod
    def submit_order(self, order: OrderIn) -> dict:
        pass

    @staticmethod
    @abstractmethod
    def order_to_status(order: OrderIn) -> dict:
        pass

    @abstractmethod
    def store_items_in_status(
        self, customer_id: str, order: str, ordered: dt.datetime, ticket_id: int, items: list[dict]
    ) -> list[Base]:
        pass

    @staticmethod
    def _fill_in_sample_ids(samples: list[dict], lims_map: dict, id_key: str = "internal_id"):
        """Fill in LIMS sample ids."""
        for sample in samples:
            LOG.debug(f"{sample['name']}: link sample to LIMS")
            if not sample.get(id_key):
                internal_id = lims_map[sample["name"]]
                LOG.info(f"{sample['name']} -> {internal_id}: connect sample to LIMS")
                sample[id_key] = internal_id

    def _add_missing_reads(self, samples: list[Sample]):
        """Add expected reads/reads missing."""
        for sample_obj in samples:
            LOG.info(f"{sample_obj.internal_id}: add missing reads in LIMS")
            target_reads = sample_obj.application_version.application.target_reads / 1000000
            self.lims.update_sample(sample_obj.internal_id, target_reads=target_reads)



import fnmatch
from pathlib import Path
from typing import Callable

import pytest

from cg.constants import FileExtensions
from cg.meta.backup.backup import BackupAPI
from cg.meta.backup.pdc import PdcAPI
from cg.meta.encryption.encryption import EncryptionAPI
from cg.meta.tar.tar import TarAPI
from cg.models.cg_config import PDCArchivingDirectory
from cg.models.cg_config import CGConfig
from cg.store.store import Store


@pytest.fixture
def mock_pdc_query_method(archived_flow_cells: list[str]) -> Callable:
    """Returns a mock method mimicking the pattern search made by the dsmc q archive command."""

    def mock_method(search_pattern: str) -> list[str] | None:
        if match := fnmatch.filter(archived_flow_cells, search_pattern):
            return match

    return mock_method


@pytest.fixture
def dsmc_q_archive_output() -> list[str]:
    output: str = """IBM Tivoli Storage Manager
Command Line Backup-Archive Client Interface
  Client Version 7, Release 1, Level 4.0
  Client date/time: 09/22/2023 10:42:42
(c) Copyright by IBM Corporation and other(s) 1990, 2015. All Rights Reserved.

Node Name: HASTA.SCILIFELAB.SE_CLINICAL
Session established with server BLACKHOLE: Linux/x86_64
  Server Version 8, Release 1, Level 9.300
  Server date/time: 09/22/2023 10:42:42  Last access: 09/22/2023 10:42:26

Accessing as node: SLLCLINICAL
             Size  Archive Date - Time    File - Expires on - Description
             ----  -------------------    -------------------------------
           607  B  04/07/2019 07:30:15    /home/hiseq.clinical/ENCRYPT/190329_A00689_0018_AHVKJCDRXX.key.gpg Never Archive Date: 04/07/2019
 1,244,997,334 KB  04/07/2019 04:00:05    /home/hiseq.clinical/ENCRYPT/190329_A00689_0018_AHVKJCDRXX.tar.gz.gpg Never Archive Date: 04/07/2019"""
    return output.splitlines()


@pytest.fixture
def archived_flow_cells(pdc_archiving_directory: PDCArchivingDirectory) -> list[str]:
    """Returns a list of archived flow cells."""
    return [
        f"{pdc_archiving_directory.current}/new_flow_cell{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}",
        f"{pdc_archiving_directory.nas}/old_flow_cell{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}",
        f"{pdc_archiving_directory.pre_nas}/ancient_flow_cell{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}",
    ]


@pytest.fixture
def backup_api(cg_context: CGConfig, illumina_flow_cells_directory: Path) -> BackupAPI:
    """Return a BackupAPI instance."""
    encryption_api: EncryptionAPI = EncryptionAPI(
        binary_path=cg_context.encryption.binary_path, dry_run=True
    )
    store: Store = cg_context.status_db
    tar_api: TarAPI = TarAPI(binary_path=cg_context.tar.binary_path, dry_run=True)
    pdc_api: PdcAPI = PdcAPI(binary_path=cg_context.pdc.binary_path, dry_run=True)

    pdc_archiving_directory: PDCArchivingDirectory = cg_context.backup.pdc_archiving_directory
    _backup_api: BackupAPI = BackupAPI(
        encryption_api=encryption_api,
        status=store,
        tar_api=tar_api,
        pdc_api=pdc_api,
        pdc_archiving_directory=pdc_archiving_directory,
        flow_cells_dir=illumina_flow_cells_directory,
        dry_run=True,
    )
    return _backup_api


@pytest.fixture
def spring_file_path() -> Path:
    """Return spring file path"""
    return Path("/path/to/spring_file.spring")


@pytest.fixture
def backup_file_path() -> str:
    """Return path to a file used in the backup process"""
    return "/path/to/backup_file.extension"


@pytest.fixture
def archived_flow_cell() -> Path:
    """Path of archived flow cell"""
    return Path("/path/to/archived/flow_cell.tar.gz.gpg")


@pytest.fixture
def archived_key() -> Path:
    """Path of archived key"""
    return Path("/path/to/archived/encryption_key.key.gpg")



"""Fixtures for flow cell objects."""

from pathlib import Path

import pytest

from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData

# Functional flow cell runs


@pytest.fixture(scope="module")
def hiseq_x_single_index_flow_cell(
    hiseq_x_single_index_flow_cell_dir: Path,
) -> FlowCellDirectoryData:
    """Return a single-index HiSeqX flow cell."""
    return FlowCellDirectoryData(flow_cell_path=hiseq_x_single_index_flow_cell_dir)


@pytest.fixture(scope="module")
def hiseq_x_dual_index_flow_cell(
    hiseq_x_dual_index_flow_cell_dir: Path,
) -> FlowCellDirectoryData:
    """Return a dual-index HiSeqX flow cell."""
    return FlowCellDirectoryData(flow_cell_path=hiseq_x_dual_index_flow_cell_dir)


@pytest.fixture(scope="module")
def hiseq_2500_dual_index_flow_cell(
    hiseq_2500_dual_index_flow_cell_dir: Path,
) -> FlowCellDirectoryData:
    """Return a dual-index HiSeq2500 flow cell."""
    return FlowCellDirectoryData(flow_cell_path=hiseq_2500_dual_index_flow_cell_dir)


@pytest.fixture(scope="module")
def hiseq_2500_custom_index_flow_cell(
    hiseq_2500_custom_index_flow_cell_dir: Path,
) -> FlowCellDirectoryData:
    """Return a custom-index HiSeq2500 flow cell."""
    return FlowCellDirectoryData(flow_cell_path=hiseq_2500_custom_index_flow_cell_dir)


@pytest.fixture
def novaseq_6000_pre_1_5_kits_flow_cell(
    illumina_flow_cells_directory: Path,
) -> FlowCellDirectoryData:
    """Return a Novaseq6000 flow cell with index settings pre 1.5 kits."""
    return FlowCellDirectoryData(
        Path(illumina_flow_cells_directory, "190927_A00689_0069_BHLYWYDSXX")
    )


@pytest.fixture
def novaseq_6000_post_1_5_kits_flow_cell(
    illumina_flow_cells_directory: Path,
) -> FlowCellDirectoryData:
    """Return a Novaseq6000 flow cell with index settings post 1.5 kits."""
    return FlowCellDirectoryData(
        Path(illumina_flow_cells_directory, "230912_A00187_1009_AHK33MDRX3")
    )


@pytest.fixture
def novaseq_x_flow_cell(novaseq_x_flow_cell_dir: Path) -> FlowCellDirectoryData:
    """Return a NovaSeqX flow cell."""
    return FlowCellDirectoryData(novaseq_x_flow_cell_dir)


@pytest.fixture
def seven_canonical_flow_cells(
    hiseq_x_single_index_flow_cell: FlowCellDirectoryData,
    hiseq_x_dual_index_flow_cell: FlowCellDirectoryData,
    hiseq_2500_dual_index_flow_cell: FlowCellDirectoryData,
    hiseq_2500_custom_index_flow_cell: FlowCellDirectoryData,
    novaseq_6000_pre_1_5_kits_flow_cell: FlowCellDirectoryData,
    novaseq_6000_post_1_5_kits_flow_cell: FlowCellDirectoryData,
    novaseq_x_flow_cell: FlowCellDirectoryData,
) -> list[FlowCellDirectoryData]:
    """Return a list with the seven canonical flow cells."""
    return [
        hiseq_x_single_index_flow_cell,
        hiseq_x_dual_index_flow_cell,
        hiseq_2500_dual_index_flow_cell,
        hiseq_2500_custom_index_flow_cell,
        novaseq_6000_pre_1_5_kits_flow_cell,
        novaseq_6000_post_1_5_kits_flow_cell,
        novaseq_x_flow_cell,
    ]


# Demultiplexed runs


@pytest.fixture
def novaseqx_flow_cell_with_sample_sheet_no_fastq(
    tmp_demultiplexed_flow_cell_no_fastq_files: Path,
) -> FlowCellDirectoryData:
    """Return a flow cell from a tmp dir with a sample sheet and no sample fastq files."""
    tmp_demultiplexed_flow_cell_no_fastq_files.mkdir(parents=True, exist_ok=True)
    flow_cell = FlowCellDirectoryData(tmp_demultiplexed_flow_cell_no_fastq_files)
    sample_sheet_path = Path(flow_cell.path, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)
    flow_cell._sample_sheet_path_hk = sample_sheet_path
    return flow_cell


@pytest.fixture
def tmp_bcl_convert_flow_cell(
    tmp_flow_cell_directory_bclconvert: Path,
) -> FlowCellDirectoryData:
    """Create a flow cell object with flow cell that is demultiplexed."""
    return FlowCellDirectoryData(tmp_flow_cell_directory_bclconvert)


@pytest.fixture
def bcl_convert_flow_cell(
    bclconvert_flow_cell_dir_name: Path,
    illumina_demultiplexed_runs_directory: Path,
) -> FlowCellDirectoryData:
    """Return a flow cell object with flow cell that is demultiplexed."""
    path = Path(illumina_demultiplexed_runs_directory, bclconvert_flow_cell_dir_name)
    return FlowCellDirectoryData(flow_cell_path=path)


# Flow cell attributes


@pytest.fixture
def bcl2fastq_flow_cell_id() -> str:
    """Return flow cell id from bcl2fastq flow cell object."""
    return "HLYWYDSXX"


@pytest.fixture
def bcl_convert_flow_cell_id(novaseq_6000_post_1_5_kits_flow_cell: FlowCellDirectoryData) -> str:
    """Return flow cell id from bcl_convert flow cell object."""
    return novaseq_6000_post_1_5_kits_flow_cell.id



from typing import Any


def join_list(potential_list: Any):
    """If given a list, it is converted to a string by joining its entries.
    Else the argument is returned as is. Used here to convert any provided synopsis to a string."""
    return "".join(potential_list) if isinstance(potential_list, list) else potential_list


def convert_well(value: str):
    """Forces the format of the well position to separate rows and values with a ':', e.g. A:8, C:3 etc."""
    if not value:
        return None
    return value if ":" in value else ":".join([value[0], value[1:]])



"""Tests the cli for uploading delivery reports."""

from click.testing import CliRunner, Result

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.upload.delivery_report import upload_delivery_report_to_scout
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.models.cg_config import CGConfig


def test_delivery_report_to_scout_no_params(upload_context: CGConfig, cli_runner: CliRunner):
    """Tests the upload to Scout without specifying the case."""

    # GIVEN a MIP-DNA report api
    assert isinstance(upload_context.meta_apis.get("report_api"), MipDNAReportAPI)

    # WHEN invoking the delivery report upload without parameters
    result: Result = cli_runner.invoke(upload_delivery_report_to_scout, obj=upload_context)

    # THEN the command should fail due to a missing case ID
    assert "There are no valid cases to perform delivery report actions" in result.output
    assert result.exit_code == EXIT_FAIL


def test_delivery_report_to_scout(
    upload_context: CGConfig,
    cli_runner: CliRunner,
    upload_report_hk_api: HousekeeperAPI,
    case_id: str,
):
    """Tests the upload to Scout of a MIP DNA delivery report."""

    # GIVEN a Housekeeper context with a delivery report file that is ready for upload
    upload_context.housekeeper_api_ = upload_report_hk_api

    # WHEN uploading the delivery report
    result: Result = cli_runner.invoke(
        upload_delivery_report_to_scout, [case_id], obj=upload_context
    )

    # THEN check that the command exits with success and that the mock file has been uploaded to Scout
    assert result.exit_code == EXIT_SUCCESS
